from spacy.tokens import Span

from conspiracies.docprocessing.coref import CoreferenceComponent  # noqa F401

from .utils import nlp_da, nlp_da_w_coref  # noqa F401


def test_coref_clusters(nlp_da_w_coref):  # noqa F811
    text = (
        "Aftalepartierne bag Rammeaftalen om plan for genåbning af Danmark blev i"
        + " foråret 2021 enige om at nedsætte en ekspertgruppe, der fik til opgave at "
        + "komme med input til den langsigtede strategi for håndtering af "
        + "coronaepidemien i Danmark. Ekspertgruppen er nu klar med sin rapport."
    )

    doc = nlp_da_w_coref(text)
    # test attributes is set as intended
    assert isinstance(doc._.coref_clusters, list)
    for sent in doc.sents:
        assert isinstance(sent._.coref_clusters, list)
        if sent._.coref_clusters != []:
            assert isinstance(sent._.coref_clusters[0][1], Span)


def test_resolve_coref(nlp_da_w_coref):  # noqa F811
    resolve_coref_text = (
        "Aftalepartierne bag Rammeaftalen om plan for genåbning af Danmark blev i"
        + " foråret 2021 enige om at nedsætte en ekspertgruppe, en ekspertgruppe fik "
        + "til opgave at komme med input til den langsigtede strategi for håndtering "
        + "af coronaepidemien i Danmark. en ekspertgruppe er nu klar med en "
        + "ekspertgruppe rapport."
    )

    resolve_coref_spans = [
        "Aftalepartierne bag Rammeaftalen om plan for genåbning af Danmark blev i "
        + "foråret 2021 enige om at nedsætte en ekspertgruppe, en ekspertgruppe fik "
        + "til opgave at komme med input til den langsigtede strategi for håndtering "
        + "af coronaepidemien i Danmark.",
        "en ekspertgruppe er nu klar med en ekspertgruppe rapport.",
    ]

    doc = nlp_da_w_coref(resolve_coref_text)
    # test for doc
    assert doc._.resolve_coref == resolve_coref_text

    # test for spans
    for i, sent in enumerate(doc.sents):
        if sent._.coref_clusters != []:
            assert sent._.resolve_coref == resolve_coref_spans[i]



import time
import urllib.request
from pathlib import Path
from typing import List, Optional, Tuple

import torch
from torch.utils.data import DataLoader

from .dataset import EvalDataset
from .extract import extract_to_dict
from .other import utils

DEFAULT_MODEL_DIR = Path(Path.home(), ".relation_model")


class KnowledgeTriplets:
    def _load_model(self, path):
        model = utils.get_models(
            bert_config=self._bert_config,
            pred_n_labels=3,
            arg_n_labels=9,
            n_arg_heads=8,
            n_arg_layers=4,
            pos_emb_dim=64,
            use_lstm=False,
            device=self._device,
        )

        model.load_state_dict(
            torch.load(path, map_location=torch.device(self._device)),
            strict=False,
        )
        model.zero_grad()
        model.eval()

        return model

    def _prepare_model(self, model_path: str = None):  # type: ignore
        if model_path is None:
            if not DEFAULT_MODEL_DIR.exists():
                DEFAULT_MODEL_DIR.mkdir()

            model_path = DEFAULT_MODEL_DIR / "relation_model_v01.bin"

            if model_path.exists():
                return self._load_model(path=model_path)
            print(f"Downloading model to {model_path}...")
            urllib.request.urlretrieve(
                url="https://sciencedata.dk//shared/81ee2688645634814152e3965e74b7f7?download",
                filename=model_path,
            )

        return self._load_model(path=model_path)

    def __init__(
        self,
        model_path: Optional[str] = None,
        batch_size: int = 64,
        max_len: int = 64,
        # it should be one if device is CPU otherwise torch is competing for cpus
        num_workers: int = 1,
        pin_memory: bool = True,
        device: Optional[str] = None,
    ):
        """A class for extracting triplets from a given text document.

        Example:
        >>> re = KnowledgeTriplets(model_path = None)
        >>> extracted_triplets = re.extract_relations(list_of_sent)

        Args:
            model_path: Path to a model (str, optional).

        Attributes:
            batch_size: An integer indicating the number of samples that will be
                propogated through the network.
            max_len: An integer defining the maximum sentence (vector?) length.
            num_workers: An integer which controls the number of worker threads which
                performe simultaneous training of a model.
            pin_memory: A boolean indicating if the fetched data Tensors should be put
                in pinned memory.
            bert_config: A string defining the configuration type.
            device: A boolean indicating whether to train a model on GPU or CPU.
        """
        self._bert_config = "bert-base-multilingual-cased"
        if not device:
            self._device = torch.device(
                "cuda:0" if torch.cuda.is_available() else "cpu",
            )
        else:
            self._device = device  # type: ignore
        self._batch_size = batch_size
        self._max_len = max_len
        self._num_workers = num_workers
        self._args = {"bert_config": self._bert_config, "device": self._device}
        self._bert_model = self._prepare_model(model_path)  # type: ignore
        self._pin_memory = pin_memory

    def _prepare_data(self, sents: List[str]):
        dataset = EvalDataset(sents, self._max_len, self._bert_config)
        test_loader = DataLoader(
            dataset,
            batch_size=self._batch_size,
            num_workers=self._num_workers,
            pin_memory=self._pin_memory,
            shuffle=False,
        )
        return test_loader

    def extract_relations(self, text: List[str], verbose: bool = False) -> List[Tuple]:
        if verbose:
            start = time.time()
        prepared_sent = self._prepare_data(sents=text)
        extractions = extract_to_dict(self._args, self._bert_model, prepared_sent)
        if verbose:
            print("TIME: ", time.time() - start)
        return extractions



import json
import logging
from typing import Iterable

from conspiracies.common.fileutils import iter_lines_of_files
from conspiracies.document import Document
from conspiracies.preprocessing.create_context_threads import context_window_thread
from conspiracies.preprocessing.preprocessor import Preprocessor

# TODO: Due to the way contexts are found and handled which requires all tweets to be
#  loaded in memory, this is not gonna scale very well for large datasets.


class TweetsPreprocessor(Preprocessor):

    def __init__(
        self,
        metadata_fields: Iterable[str] = ("*",),
        context_length: int = 2,
    ):
        super().__init__(metadata_fields=metadata_fields)
        self.context_length = context_length

    def _do_preprocess_docs(self, glob_pattern: str):
        lines = iter_lines_of_files(glob_pattern)
        tweets = [json.loads(line) for line in lines]

        doc_ids = set()
        tweets_with_context = context_window_thread(tweets, self.context_length)
        for tweet_with_context in tweets_with_context:
            context_tweets, tweet = tweet_with_context[:-1], tweet_with_context[-1]
            doc_id = tweet["id"]
            if doc_id in doc_ids:
                logging.debug("Skipping duplicate %s", doc_id)
                continue
            doc_ids.add(doc_id)
            if (
                "referenced_tweets" in tweet
                and tweet["referenced_tweets"][0]["type"] == "retweeted"
            ):
                logging.debug("Skipping retweet %s", doc_id)
                continue
            metadata = {k: v for k, v in tweet.items() if k != "text"}
            text = tweet["text"]
            context = "\n".join(t["text"] for t in context_tweets)
            yield Document(id=doc_id, metadata=metadata, text=text, context=context)



import csv
from typing import Iterable, Iterator

from conspiracies.common.fileutils import iter_lines_of_files
from conspiracies.document import Document
from conspiracies.preprocessing.preprocessor import Preprocessor


class CsvPreprocessor(Preprocessor):

    def __init__(
        self,
        id_column: str = None,
        text_column: str = None,
        context_column: str = None,
        delimiter=",",
        metadata_fields: Iterable[str] = ("*",),
    ):
        self.id_column = id_column
        self.text_column = text_column
        self.context_column = context_column
        self.non_metadata_columns = {
            self.id_column,
            self.text_column,
            self.context_column,
        }
        self.delimiter = delimiter
        super().__init__(metadata_fields=metadata_fields)

    def _read_lines(self, lines: Iterable[str]) -> Iterator[str]:
        reader = csv.DictReader(lines, delimiter=self.delimiter)
        for row in reader:  # type: dict
            id_ = row[self.id_column]
            text = row[self.text_column]
            context = row[self.context_column] if self.context_column else None
            metadata = {
                k: v for k, v in row.items() if k not in self.non_metadata_columns
            }
            yield Document(
                id=id_,
                text=text,
                context=context,
                metadata=metadata,
            )

    def _do_preprocess_docs(self, input_path: str) -> Iterator[str]:
        lines = iter_lines_of_files(input_path)
        return self._read_lines(lines)



import pytest
import spacy

from .test_data.prompt_data import load_gold_triplets


@pytest.fixture
def nlp_en():
    return spacy.load("en_core_web_sm")


@pytest.fixture
def nlp_da_w_coref(nlp_da):
    nlp_da.add_pipe("allennlp_coref")
    return nlp_da


@pytest.fixture
def nlp_da():
    return spacy.load("da_core_news_sm")


@pytest.fixture
def docs_with_triplets():
    return load_gold_triplets()



import pytest

from conspiracies.document import Document
from conspiracies.preprocessing.csv import CsvPreprocessor


@pytest.fixture
def csv_with_headers():
    lines = [
        "id,text,context,metadata1,metadata2",
        "1,some text,a bit of context,one,two",
    ]
    return lines


@pytest.fixture
def expected_docs():
    return [
        Document(
            id="1",
            text="some text",
            context="a bit of context",
            metadata={"metadata1": "one", "metadata2": "two"},
        ),
    ]


def test_read_lines(csv_with_headers, expected_docs):
    preprocessor = CsvPreprocessor(
        id_column="id",
        text_column="text",
        context_column="context",
    )
    # NOTE: metadata fields are not filtered in this method, so all other fields will be
    #   included as metadata
    assert list(preprocessor._read_lines(csv_with_headers)) == expected_docs



import spacy
from spacy.tokens import Doc, Span

from conspiracies.docprocessing.headwordextraction.headwordextraction_component import (
    contains_ents,
)


def test_ents_filter():
    nlp = spacy.blank("en")
    nlp.add_pipe("heads_extraction")
    doc = Doc(nlp.vocab, words="Mette Frederiksen is the Danish Politician".split())
    doc.set_ents([Span(doc, 0, 2, "PERSON")])

    assert contains_ents(doc[:]) is True
    assert contains_ents(doc[1:2]) is True
    assert contains_ents(doc[2:4]) is False



import json
from typing import Tuple, List, Dict, Optional, Union
import os
import spacy
from umap import UMAP
from hdbscan import HDBSCAN
from sentence_transformers import SentenceTransformer
from stop_words import get_stop_words
from sklearn.preprocessing import StandardScaler
from numpy import ndarray
from collections import Counter
import random
import argparse


def read_txt(path: str):
    with open(path, mode="r", encoding="utf8") as f:
        lines = f.read().splitlines()
    return lines


def triplet_from_line(line: str) -> Union[Tuple[str, str, str], None]:
    """Converts a line from a txt file to a triplet.

    Lines that are not exactly three elements are ignored.
    All elements in the triplet are stripped of whitespace and lowercased.
    Args:
        line: Line from a txt file
    Returns:
        triplet: Triplet
    """
    as_list = line.split(", ")
    if len(as_list) != 3:
        return None
    if line in ["Subject, Predicate, Object", "---, ---, ---"]:
        return None
    return tuple(map(str.strip, map(str.lower, as_list)))  # type: ignore


def filter_triplets_with_stopwords(
    triplets: List[Tuple[str, str, str]],
    stopwords: List[str],
    soft: bool = True,
) -> List[Tuple[str, str, str]]:
    """Filters triplets that contain a stopword.

    Args:
        triplets: List of triplets. A triplet is a tuple of
            three strings.
        stopwords: List of stopwords
        soft: If True, only the subject and object are checked for stopwords. If
            False, the whole triplet is checked.
    Returns:
        filtered_triplets: List of triplets without
        stopwords
    """
    filtered_triplets = []
    if soft:
        for triplet in triplets:
            subject, predicate, obj = triplet
            if subject not in stopwords and obj not in stopwords:
                filtered_triplets.append(triplet)
    else:
        for triplet in triplets:
            if not any(stopword in triplet for stopword in stopwords):
                filtered_triplets.append(triplet)
    return filtered_triplets


def load_triplets(
    file_path: str,
    soft_filtering: bool = True,
    shuffle: bool = True,
) -> Tuple[list, list, list, list]:
    """Loads triplets from a file and filters them.

    Args:
        file_name: Name of the file to load triplets from
        soft_filtering: Whether to use soft filtering or not
        shuffle: Whether to shuffle the triplets or not
    Returns:
        subjects: List of subjects
        predicates: List of predicates
        objects: List of objects
        filtered_triplets: List of filtered triplets
    """
    triplets_list: List[Tuple[str, str, str]] = []
    data = read_txt(file_path)
    triplets_list = [
        triplet_from_line(line)
        for line in data
        if triplet_from_line(line)  # type: ignore
    ]
    filtered_triplets = filter_triplets_with_stopwords(
        triplets_list,
        get_stop_words("danish"),
        soft=soft_filtering,
    )

    if shuffle:
        random.shuffle(filtered_triplets)

    subjects = [
        triplet[0]
        for triplet in filtered_triplets
        if triplet[0] not in ["Subject", "---"]
    ]
    predicates = [
        triplet[1]
        for triplet in filtered_triplets
        if triplet[0] not in ["Predicate", "---"]
    ]
    objects = [
        triplet[2]
        for triplet in filtered_triplets
        if triplet[0] not in ["Object", "---"]
    ]
    return subjects, predicates, objects, filtered_triplets


def freq_of_most_frequent(list_of_strings: List[str]) -> Tuple[str, float]:
    """Calculates the frequency of the most frequent element in a list of
    strings.

    Frequency is measured as how much of the list the most frequent element takes up,
        percentage-wise.
    Args:
        list_of_strings: List of strings to find the most frequent element
        in
    Returns:
        most_common_string (str), percentage: Frequency of the most frequent
        element, its percentage
    """
    most_common = Counter(list_of_strings).most_common(1)[0]
    most_common_string = most_common[0]
    percentage = most_common[1] / len(list_of_strings)
    return most_common_string, percentage


def most_frequent_token(list_of_strings: List[str], nlp) -> Tuple[str, float]:
    """Finds the most frequent token in a list of strings.

    Args:
        list_of_strings: List of strings to find the most frequent token in
        nlp: Spacy language model
    Returns:
        most_common_token: Most frequent token
    """
    token_list = []
    for string in list_of_strings:
        doc = nlp(string)
        for token in doc:
            token_list.append(token.text)
    most_common_token, percentage = freq_of_most_frequent(token_list)
    return most_common_token, percentage


def get_cluster_label(
    cluster: List[Tuple[str, int]],
    nlp,
    first_cutoff: float = 0.8,
    min_cluster_length: int = 10,
    second_cutoff: float = 0.3,
) -> Union[str, None]:
    """Finds the most frequent token in a cluster.

    Args:
        cluster: Cluster of elements
        nlp: Spacy language model
        certain_cutoff: Minimum percentage of the most frequent token.
            All clusters that have a most frequent token at least as frequent as this
            value gets that token as label.
    Returns:
        label: The cluster label according to the rules
    """
    cluster_strings = [element[0] for element in cluster]
    most_common, percentage = freq_of_most_frequent(cluster_strings)

    # If the most frequent token is frequent enough, use it as label
    if percentage >= first_cutoff:
        return most_common

    else:
        # If the cluster is not clearly defined and it is short,
        # return None to indicate it should be removed
        if len(cluster) < min_cluster_length:
            return None

        # Clusters that are large and relatively clearly defined
        if percentage >= second_cutoff:
            return most_common

        # If the cluster is not clearly defined, find the most frequent token
        most_common_token, percentage = most_frequent_token(cluster_strings, nlp)

        # The most frequent token must be frequent enough to be used as label,
        # otherwise return None to indicate it should be removed
        if percentage >= 0.2:  # TODO: Make this a parameter
            return most_common_token
        else:
            return None


def cluster_dict(topic_labels: ndarray, input_list: List[str]) -> Dict[int, List[str]]:
    """Creates a dictionary containing all elements in a cluster from a
    BERTopic model.

    Only clusters with at least `cutoff` elements.
    Args:
        topic_model: BERTopic model
        input_list: The list of string that were used to create the BERTopic
        model
        cutoff: Minimum number of elements in a cluster
    Returns:
        cluster_dict: Dictionary of clusters
    """
    assert len(topic_labels) == len(
        input_list,
    ), "Length of topic labels and input list must be equal"
    # topic_info = topic_model.get_topic_info()
    # relevant_topics = topic_info.loc[topic_info["Count"] >= cutoff]
    cluster_dict: Dict[int, list] = {i: [] for i in range(max(topic_labels) + 1)}
    topic_tuples = zip(topic_labels, input_list)
    for index, (topic_n, element) in enumerate(topic_tuples):
        if topic_n != -1:
            cluster_dict[topic_n].append((element, index))
    return cluster_dict


def label_clusters(
    cluster_dict,
    nlp,
    first_cutoff: float = 0.8,
    min_cluster_length: int = 10,
    second_cutoff: float = 0.3,
    predicates: bool = False,
):
    """Labels clusters according to the rules in `get_cluster_label`.
    Args:
        cluster_dict: Dictionary of clusters to label
        nlp: Spacy language model to use for tokenization of less defined
            clusters
        first_cutoff: Minimum percentage of the most frequent element.
            All clusters that have a most frequent element at least as frequent as this
            value gets that element as label.
        min_cluster_length: Minimum number of elements in a cluster
        second_cutoff: Minimum percentage of the most frequent element.
            This cutoff is only used if the cluster has a less clearly defined label,
            but is still large enough.

    Returns:
        dict_with_labels: Dictionary of clusters with
        labels
            The keys are the labels, the values are dictionaries with the keys
                "cluster" (final elements in the cluster), and
                "n_elements" (number of elements in the final cluster)
    """
    if predicates:
        min_cluster_length = 1
        second_cutoff = 0.0
    dict_with_labels: Dict[str, dict] = {}
    # Get the label for each cluster
    for cluster in cluster_dict.values():
        cluster_label = get_cluster_label(
            cluster,
            nlp,
            first_cutoff,
            min_cluster_length,
            second_cutoff,
        )
        if cluster_label:  # If the cluster is not None
            if (
                cluster_label in dict_with_labels.keys()
            ):  # If the label already exists, merge
                dict_with_labels[cluster_label]["cluster"].extend(cluster)
                dict_with_labels[cluster_label]["n_elements"] += len(cluster)
            else:  # If the label does not exist, create a new entry
                dict_with_labels[cluster_label] = {}
                dict_with_labels[cluster_label]["cluster"] = cluster
                dict_with_labels[cluster_label]["n_elements"] = len(cluster)

    # Remove clusters that are too small even after merging identical clusters
    clusters_to_keep = {
        label: content
        for label, content in dict_with_labels.items()
        if content["n_elements"] > min_cluster_length
    }
    return clusters_to_keep


def embed_and_cluster(
    list_to_embed: List[str],
    embedding_model: str = "vesteinn/DanskBERT",
    n_dimensions: int = 40,
    n_neighbors: int = 15,
    min_cluster_size: int = 5,
    min_samples: int = 3,
    min_topic_size: int = 10,
    predicates: bool = False,
):
    """Embeds and clusters a list of strings.

    Args:
        list_to_embed: List of strings to embed and cluster
        n_dimensions: Number of dimensions to reduce the embedding space to
        n_neighbors: Number of neighbors to use for UMAP
        min_cluster_size: Minimum cluster size for HDBscan
        min_samples: Minimum number of samples for HDBscan
        min_topic_size: Minimum number of elements in a cluster
    Returns:
        clusters: Dictionary of clusters with labels
            The keys are the labels, the values are dictionaries with the keys
                "cluster" (final elements in the cluster), and
                "n_elements" (number of elements in the final cluster)
    """

    embedding_model = SentenceTransformer(embedding_model)

    # Embed and reduce embdding space
    print("Embedding and reducing embedding space")
    embeddings = embedding_model.encode(list_to_embed)  # type: ignore
    scaled_embeddings = StandardScaler().fit_transform(embeddings)
    reducer = UMAP(n_components=n_dimensions, n_neighbors=n_neighbors)
    reduced_embeddings = reducer.fit_transform(scaled_embeddings)

    # Cluster with HDBscan
    print("Clustering")
    hdbscan_model = HDBSCAN(min_cluster_size=min_cluster_size, min_samples=min_samples)
    hdbscan_model.fit(reduced_embeddings)
    hdbscan_labels = hdbscan_model.labels_
    assert len(hdbscan_labels) == len(
        list_to_embed,
    ), "Length of hdbscan labels and input list must be equal"
    clusters = cluster_dict(hdbscan_labels, list_to_embed)

    # Label and prune clusters
    print("Labeling clusters")
    nlp = spacy.load("da_core_news_sm")
    labeled_clusters = label_clusters(
        clusters,
        nlp,
        min_cluster_length=min_topic_size,
        predicates=predicates,
    )

    return labeled_clusters


def create_nodes_and_edges(
    subj_obj_clusters: Dict[str, Dict[str, List[str]]],
    predicate_clusters: Dict[str, Dict[str, List[str]]],
    n_elements: int,
    no_predicate_filler: str = "",
    save: Optional[Union[bool, str]] = False,
):
    """Creates nodes and edges from clusters of subjects, objects and predicates.
    Args:
        subj_obj_clusters : Dictionary of clusters
            with labels
            The keys are the labels, the values are dictionaries with the keys
                "cluster" (final elements in the cluster), and
                "n_elements" (number of elements in the final cluster)
        predicate_clusters : Dictionary of clusters
            with labels
            The keys are the labels, the values are dictionaries with the keys
                "cluster" (final elements in the cluster), and
                "n_elements" (number of elements in the final cluster)
        no_predicate_filler : String to use as filler for predicates that do not
            have a cluster
        save : If a string, nodes and edges are saved to a
            json file. If False, does not save.

    Returns:
        nodes, edges: nodes and edges for the graph
    """

    labelled_subjects = {i: "" for i in range(0, n_elements)}
    labelled_objects = {i: "" for i in range(0, n_elements)}

    for label, content in subj_obj_clusters.items():
        cluster = content["cluster"]
        for _, index in cluster:  # type: ignore
            if index < n_elements:  # type: ignore
                labelled_subjects[index] = label  # type: ignore
            else:
                labelled_objects[index - n_elements] = label  # type: ignore
    labelled_subjects = {  # type: ignore
        i: label
        for i, label in labelled_subjects.items()
        if label != ""  # type: ignore
    }
    labelled_objects = {
        i: label for i, label in labelled_objects.items() if label != ""
    }

    labelled_predicates = {i: "" for i in range(0, n_elements)}
    for label, content in predicate_clusters.items():
        cluster = content["cluster"]
        for _, index in cluster:  # type: ignore
            labelled_predicates[index] = label  # type: ignore
    labelled_predicates = {
        i: label for i, label in labelled_predicates.items() if label != ""
    }

    nodes = []
    edges = []
    for s_index, subject in labelled_subjects.items():
        if s_index in labelled_objects.keys():
            nodes.append((subject, labelled_objects[s_index]))
            if s_index in labelled_predicates.keys():
                edges.append(labelled_predicates[s_index])
            else:
                edges.append(no_predicate_filler)

    if save:
        with open(save, "w") as f:
            json.dump({"nodes": nodes, "edges": edges}, f)
    return nodes, edges


def main(
    path: str,
    embedding_model: str,
    dim=40,
    n_neighbors=15,
    min_cluster_size=5,
    min_samples=3,
    min_topic_size=20,
    save: bool = False,
):
    # Load triplets
    print("Loading triplets")
    subjects, predicates, objects, filtered_triplets = load_triplets(
        path,
        soft_filtering=True,
        shuffle=True,
    )

    if save and not isinstance(save, str):
        save = path.replace(
            "triplets.txt",
            f"{embedding_model}_dim={dim}_neigh={n_neighbors}"
            f"_clust={min_cluster_size}_samp={min_samples}_nodes_edges.json",
        )  # type: ignore

    model = (
        "vesteinn/DanskBERT"
        if embedding_model == "danskBERT"
        else "sentence-transformers/paraphrase-multilingual-mpnet-base-v2"
    )

    print(
        f"Dimensions: {dim}, neighbors: {n_neighbors}, min cluster size: "
        f"{min_cluster_size}, samples: {min_samples}, min topic size: {min_topic_size}",
    )
    print("\n_________________\n")
    print("Embedding and clustering predicates")
    # For predicate, we wanna keep all clusters -> min_topic_size=1
    predicate_clusters = embed_and_cluster(
        list_to_embed=predicates,
        embedding_model=model,
        n_dimensions=dim,
        n_neighbors=n_neighbors,
        min_cluster_size=min_cluster_size,
        min_samples=min_samples,
        min_topic_size=1,
        predicates=True,
    )

    print("\n_________________\n")
    print("Embedding and clustering subjects and objects together")
    subj_obj = subjects + objects
    subj_obj_clusters = embed_and_cluster(
        list_to_embed=subj_obj,
        embedding_model=model,
        n_dimensions=dim,
        n_neighbors=n_neighbors,
        min_cluster_size=min_cluster_size,
        min_samples=min_samples,
        min_topic_size=min_topic_size,
    )

    # Create nodes and edges
    print("Creating nodes and edges")

    assert (
        len(subjects) == len(objects) == len(predicates)
    ), "Subjects, objects and predicates must have the same length"
    nodes, edges = create_nodes_and_edges(
        subj_obj_clusters,
        predicate_clusters,
        len(subjects),
        save=save,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-e",
        "--event",
        type=str,
        help="Event to cluster. Must include name of source folder (newspapers or "
        "twitter) and event",
    )
    parser.add_argument(
        "-emb",
        "--embedding_model",
        type=str,
        default="paraphrase",
        help="""Which embedding model to use, default is paraphrase. 
        The other option is danskBERT""",
    )
    parser.add_argument(
        "-dim",
        "--n_dimensions",
        type=int,
        default=40,
        help="Number of dimensions to reduce the embedding space to",
    )
    parser.add_argument(
        "-neigh",
        "--n_neighbors",
        type=int,
        default=15,
        help="Number of neighbors to use for UMAP",
    )
    parser.add_argument(
        "-min_clust",
        "--min_cluster_size",
        type=int,
        default=5,
        help="Minimum cluster size for HDBscan",
    )
    parser.add_argument(
        "-min_samp",
        "--min_samples",
        type=int,
        default=3,
        help="Minimum number of samples for HDBscan",
    )
    parser.add_argument(
        "-save",
        "--save",
        type=bool,
        default=False,
        help="whether or not to save nodes and edges to json file",
    )

    args = parser.parse_args()
    path = os.path.join(args.event, "triplets.txt")
    main(
        path,
        embedding_model=args.embedding_model,
        dim=args.n_dimensions,
        n_neighbors=args.n_neighbors,
        min_cluster_size=args.min_cluster_size,
        min_samples=args.min_samples,
        save=args.save,
    )



import catalogue
from confection import registry

registry.prompt_templates = catalogue.create(  # type: ignore
    "conspiracies",
    "prompt_templates",
)
# a prompt template class object

registry.split_doc_functions = catalogue.create(  # type: ignore
    "conspiracies",
    "split_functions",
)
# split functions:
# Callable[[Doc], List[Span]]

registry.prompt_apis = catalogue.create("conspiracies", "prompt_apis")  # type: ignore
# prompt APIs:
# Callable[[PromptTemplate, str, str, Dict[Any, Any]], Callable[[str], str]]
# corresponding to the prompt_template, api_key, model_name, api_kwargs and return
# a function which takes in a target string and returns a output response (str).



"""Pipeline for headwords/entities extractions and frequency count."""

import spacy
from tqdm import tqdm

from conspiracies.docprocessing.headwordextraction import contains_ents


def main():
    nlp = spacy.load("en_core_web_lg")

    test_sents = ["Mette Frederiksen is the Danish politician."]

    config = {"confidence_threshold": 2.7, "model_args": {"batch_size": 10}}
    nlp.add_pipe("relation_extractor", config=config)
    nlp.add_pipe("heads_extraction")

    pipe = nlp.pipe(test_sents)

    heads_spans = []
    ents_spans = []

    for d in tqdm(pipe):
        for span in d._.relation_head:
            heads_spans.append(span._.most_common_ancestor)
            if span.ents:
                ents_spans.append(list(span.ents))
        for span in d._.relation_tail:
            heads_spans.append(span._.most_common_ancestor)
            if span.ents:
                ents_spans.append(list(span.ents))

    # Filter out headwords without an entity type
    filtered_heads = list(filter(contains_ents, heads_spans))

    # Flatten the list with ents
    flat_ents_list = [ent for sublist in ents_spans for ent in sublist]


if __name__ == "__main__":
    main()



"""A relation extraction component for spaCy using prompt-based relation
extraction."""

from typing import Any, Dict, Iterable, List, Optional

import numpy as np
from spacy.language import Language
from spacy.tokens import Doc
from spacy.training.example import Example

from conspiracies.registry import registry
from conspiracies.docprocessing.relationextraction.data_classes import (
    DocTriplets,
    SpanTriplet,
    install_extensions,
)
from .prompt_apis import create_openai_chatgpt_prompt_api  # noqa: F401


def score_open_relations(examples: Iterable[Example]) -> Dict[str, Any]:
    """Score the predicted relations against the gold relations."""
    keys = [
        "exact_span_match",
        "exact_string_match",
        "normalized_span_overlap",
        "normalized_string_overlap",
    ]
    hits = {key: 0 for key in keys}
    n_pred = 0
    n_ref = 0
    sample_scores = []

    for example in examples:
        gold_doc_triplets = example.reference._.relation_triplets
        pred_doc_triplets = example.predicted._.relation_triplets

        _score = pred_doc_triplets.score_relations(gold_doc_triplets)
        sample_scores.append(_score)

        hits["exact_span_match"] += _score["exact_span_match"]
        hits["exact_string_match"] += _score["exact_string_match"]
        hits["normalized_span_overlap"] += _score["normalized_span_overlap"]
        hits["normalized_string_overlap"] += _score["normalized_string_overlap"]
        n_pred += _score["length_self"]
        n_ref += _score["length_reference"]

    scores: Dict[str, Any] = {}

    # calculate precision, recall, f1_macro
    for key in hits:
        if n_pred:
            scores[f"{key}_precision"] = hits[key] / n_pred
        else:
            scores[f"{key}_precision"] = np.nan
        if n_ref:
            scores[f"{key}_recall"] = hits[key] / n_ref
        else:
            scores[f"{key}_recall"] = np.nan

        if scores[f"{key}_precision"] == 0 and scores[f"{key}_recall"] == 0:
            scores[f"{key}_f1"] = np.nan
        else:
            scores[f"{key}_f1"] = (
                2
                * (scores[f"{key}_precision"] * scores[f"{key}_recall"])
                / (scores[f"{key}_precision"] + scores[f"{key}_recall"])
            )

    scores["sample_scores"] = sample_scores
    scores["n_predictions"] = n_pred
    scores["n_references"] = n_ref

    return scores


class PromptRelationExtractionComponent:
    """A class for extracting relations from a spaCy Doc using prompt-based
    relation extraction."""

    def __init__(
        self,
        nlp: Language,
        name: str,
        prompt_template: str,
        examples: Optional[List[List[SpanTriplet]]],
        task_description: Optional[str],
        model_name: str,
        backend: str,
        api_key: str,
        api_kwargs: Dict[Any, Any],
        split_doc_fn: Optional[str],
        force: bool,
    ):
        """Initialise components."""
        self.name = name
        self.nlp = nlp
        p_template = registry.get("prompt_templates", prompt_template)

        self.prompt_template = p_template(
            task_description=task_description,
            examples=examples,
        )

        # create prompt function using the desired API
        self.prompt_fn = registry.get("prompt_apis", backend)(
            self.prompt_template,
            api_key,
            model_name,
            api_kwargs,
        )
        self.model_name = model_name
        self.backend = backend
        self.api_key = api_key
        self.api_kwargs = api_kwargs
        if split_doc_fn is not None:
            self.split_doc_fn = registry.get("split_doc_functions", split_doc_fn)()
        else:
            self.split_doc_fn = None

        if not Doc.has_extension("relation_triplets") or force:
            install_extensions(force=force)

    def combine_docs(self, docs: List[Doc]) -> Doc:
        """Combine a list of docs into a single doc."""
        raise NotImplementedError("Combine docs not implemented yet.")

    def set_annotation(self, doc, docs: List[Doc], prompt_outputs: List[str]) -> Doc:
        """Set the annotation on the doc."""
        spantriplets = []
        for doc_span, prompt in zip(docs, prompt_outputs):
            triplets = self.prompt_template.parse_prompt(
                prompt,
                target_tweet=doc_span.text,
            )
            spantriplets.extend(triplets)

        doc._.relation_triplets = DocTriplets.from_str_triplets(
            doc,
            spantriplets,
        )
        return doc

    def score(self, examples: Iterable[Example]) -> Dict[str, Any]:
        """Score the predicted relations against the gold relations. Scores
        using the following attributes:

        - `exact_span_match`: Whether there is an exact span match of the all three of
            the triplets (1 or 0).
        - `exact_string_match`: Whether there is an exact string match of the all three
            of the triplets (1 or 0).
        - `normalized_span_overlap`: The normalized span overlap between the predicted
            and gold relations. A value of 1 indicates an exact match (0-1, exact span
            match is 1).
        - `normalized_string_token_overlap`: The normalized string token overlap between
            the predicted and gold relations. A value of 1 indicates an exact match
            (0-1, exact string match is 1).

        For each of these and `{score}.f1_macro`, `{score}.f1_micro`, `{score}.recall`,
        `{score}.precision` are calculated.

        Args:
            examples (Iterable[Example]): A list of spaCy Examples.

        Returns:
            Dict[str, Any]: A dictionary of scores.
        """
        return score_open_relations(examples)

    def __call__(self, doc: Doc):
        """Run the pipeline component."""
        # split into tweets
        if self.split_doc_fn is not None:
            doc_spans = self.split_doc_fn(doc)
        else:
            doc_spans = [doc]

        # prompt
        responses = self.prompt_fn([span.text for span in doc_spans])
        doc = self.set_annotation(doc, doc_spans, responses)
        return doc


@Language.factory(
    "conspiracies/prompt_relation_extraction",
    default_config={
        "prompt_template": "conspiracies/template_1",
        "examples": None,
        "task_description": None,
        "model_name": "text-davinci-002",
        "backend": "conspiracies/openai_gpt3_api",
        "split_doc_fn": None,
        "api_kwargs": {
            "max_tokens": 500,
            "temperature": 0.7,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
        },
        "force": True,
    },
)
def create_prompt_relation_extraction_component(
    nlp: Language,
    name: str,
    prompt_template: str,
    examples: Optional[List[List[SpanTriplet]]],
    task_description: Optional[str],
    model_name: str,
    backend: str,
    api_key: str,
    api_kwargs: Dict[Any, Any],
    split_doc_fn: Optional[str],
    force: bool,
) -> PromptRelationExtractionComponent:
    """Allows PromptRelationExtractionComponent to be added to a spaCy pipe
    using nlp.add_pipe("conspiracies/prompt_relation_extraction").

    Args:
        nlp (Language): A spaCy Language object.
        name (str): The instance name of the component in the pipeline.
        prompt_template (str): The name of the prompt template to use. Defaults to
            "conspiracies/template1". Available templates are reqistered in
            registry.prompt_templates.
        examples (List[List[SpanTriplet]]): A list of lists of SpanTriplets to use as
            examples for the prompt. Defaults to None. If None, the examples will be
            taken from the prompt template.
        task_description (str): A description of the task to be performed. Defaults to
            None. If None, the task description will be taken from the prompt template.
        model_name (str): The name of the model to use for the prompt. Defaults to
            "text-davinci-002".
        backend (Literal["openai", "huggingface"]): The backend to use for the prompt.
        force (bool): Whether to force the extension to be added to the Doc.

    Returns:
        PromptRelationExtractionComponent: A spaCy component for prompt-based relation
            extraction.

    Example:
        >>> import spacy
        >>> import conspiracies
        >>> nlp = spacy.load("en_core_web_sm")
        >>> config = {"prompt_template": "conspiracies/template1",
        ...           "examples": None,
        ...           "task_description": None,
        ...           "force": True}
        >>> nlp.add_pipe("conspiracies/prompt_relation_extraction", config=config)
        >>> doc = nlp("This is a test document.")
        >>> doc._.relation_triplet
    """

    return PromptRelationExtractionComponent(
        nlp,
        name=name,
        prompt_template=prompt_template,
        examples=examples,
        task_description=task_description,
        model_name=model_name,
        backend=backend,
        api_key=api_key,
        api_kwargs=api_kwargs,
        split_doc_fn=split_doc_fn,
        force=force,
    )



from typing import List

from spacy.language import Language
from spacy_transformers.layers.transformer_model import huggingface_tokenize

from transformers import PreTrainedTokenizer


def wordpiece_length_normalization(
    texts: List[str],
    nlp: Language,
    tokenizer: PreTrainedTokenizer,
    max_length: int = 500,
):
    """Yield a list of text split into subtext which respect the maximum
    wordpiece or sentencepiece tokens (max_length).

    Args:
        texts: (List[str]): List of strings.
        nlp (Language): A spacy Language pipeline
        tokenizer (str): The pretrained transformer tokenizer.
        max_length (int): The maximum number of tokens of the model. Defualt to 500.

    Example:
        >>> import spacy
        >>> from transformers import AutoTokenizer
        >>> nlp = spacy.load("da_core_news_lg")
        >>> short_texts = ["hello there"]
        >>> tokenizer_name = "DaNLP/da-bert-tone-subjective-objective"
        >>> tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)
        >>> norm_text = wordpiece_length_normalization(short_texts, nlp, tokenizer)
        >>> list(norm_text)  # short texts should result in the same output
        ["hello there"]
    """

    def __wordpiece_length(text: str):
        wp_tokens = huggingface_tokenize(tokenizer, [text])
        return len(wp_tokens["input_ids"][0])

    def __split_sentence(sent):
        nonlocal min_span

        doc = sent.doc
        max_span = sent.start + 1
        while max_span < sent.end:
            span_wp_length = __wordpiece_length(doc[min_span:max_span].text)
            if span_wp_length > max_length:
                if min_span + 1 == max_span:
                    raise ValueError(
                        "The following token span consist"
                        + f" of more than {max_length} word pieces.",
                    )
                yield doc[min_span:max_span].text
                min_span = max_span
            max_span += 1
        # yield remaining sentence
        if min_span < sent.end:
            yield doc[min_span : sent.end].text
        min_span = sent.end + 1

    docs = nlp.pipe(texts)
    for doc in docs:
        doc_wp_length = __wordpiece_length(doc.text)

        if doc_wp_length < max_length:
            yield doc.text
        else:  # if doc needs to be split
            min_span = 0

            for sent in doc.sents:
                span_wp_length = __wordpiece_length(doc[min_span : sent.end].text)

                if span_wp_length > max_length:
                    if min_span == sent.start:
                        # if the sentence is too long split it to longest possible
                        # substring respecting token bounderies.
                        for sub_sent in __split_sentence(sent):
                            yield sub_sent
                    # if too long yield up until sentence
                    else:
                        yield doc[min_span : sent.start].text
                        min_span = sent.start

            if min_span < sent.end:
                yield doc[min_span : sent.end].text  # yield remaining sentence



"""Extracting longest thread from twitter corpus."""

import os
from glob import glob
import ndjson
import argparse
import datetime
import re


def ndjson_gen(filepath: str):
    """Creates a generator for ndjson files and yields the tweets."""
    for file in glob(filepath):
        print("Reading file:", file)
        with open(file) as f:
            reader = ndjson.reader(f)
            for post in reader:
                if re.search("^RT", post["text"]):  # remove retweets
                    continue
                yield post


'''
### context window thread ###
def context_window_thread(tweets, context_len, start_date, extra_days):
    """
    Extracts a thread of context tweets; the last tweet in the thread is the one that the context relates to
    Args:
        tweets: list of tweets
        context_len: number of tweets to include in context
        start_date: start date of the date range
        extra_days: number of days to include from start date and after
    Returns:
        contexts: list of contexts, where each context is a list of tweets
    """
    conversations: dict = {}
    # loop over tweets and create conversation dicts
    for tweet in tweets:
        
        conv_id = tweet["conversation_id"]
        if conv_id not in conversations.keys():
            conversations[conv_id] = {tweet["id"]: tweet}

        else:
            conversations[conv_id][tweet["id"]] = tweet

    def find_reply_to_id(tweet: dict):
        """
        Find the id of the tweet this tweet is replying to
        """
        for ref in tweet["referenced_tweets"]:
            if ref["type"] == "replied_to":
                return ref["id"]

    def find_context(tweet: dict, conversation: dict, context_length: int = context_len):
        if "in_reply_to_user_id" in tweet.keys():
            if context_length == 0:
                return [tweet]
            else:
                if "referenced_tweets" not in tweet.keys():
                    #raise Warning("reply but no reference?") #googl
                    print("reply but no reference?") 
                    return [tweet]

                reply_to_id = find_reply_to_id(tweet)
                try:
                    reply_to = conversation[reply_to_id]
                except KeyError:
                    
                    return [tweet]

                context = find_context(
                    reply_to,
                    conversation,
                    context_length=context_length - 1,
                )
                context.append(tweet)
                return context
        else:
            return [tweet]
    
    contexts = []
    for tweet in tweets:
        context = find_context(tweet, conversations[tweet["conversation_id"]], context_length=context_len)
        
        if len(context) > 1:
            contexts.append(context)
        elif len(context) >= context_len+1:
            print("context length too long:", len(context))
            break
    return contexts
'''


def context_window_thread(tweets, context_len):
    """
    Extracts a thread of context tweets; the last tweet in the thread is the one that the context relates to
    Args:
        tweets: list of tweets
        context_len: number of tweets to include in context
        start_date: start date of the date range
        extra_days: number of days to include from start date and after
    Returns:
        contexts: list of contexts, where each context is a list of tweets
    """
    conversations = {}
    for tweet in tweets:
        conv_id = tweet["conversation_id"]
        conversations.setdefault(conv_id, {})
        conversations[conv_id][tweet["id"]] = tweet

    def find_context(tweet, conversation, context_length=context_len):
        if "in_reply_to_user_id" not in tweet:
            return [tweet]

        if context_length == 0 or "referenced_tweets" not in tweet:
            return [tweet]

        replied_to_refs = [
            ref for ref in tweet["referenced_tweets"] if ref["type"] == "replied_to"
        ]
        if replied_to_refs:
            reply_to_id = next(ref["id"] for ref in replied_to_refs)
            try:
                reply_to = conversation[reply_to_id]
            except KeyError:
                return [tweet]

            context = find_context(
                reply_to,
                conversation,
                context_length=context_length - 1,
            )
            context.append(tweet)
            return context
        else:
            # handle the case where no replied_to reference is found
            return [tweet]

    contexts = [
        find_context(tweet, conversations[tweet["conversation_id"]])
        for tweet in tweets
        if len(find_context(tweet, conversations[tweet["conversation_id"]])) > 1
    ]
    return contexts


def range_of_dates(start_date, extra_days):
    start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d").date()
    date_range = [start_date + datetime.timedelta(days=x) for x in range(extra_days)]
    return date_range


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Get tweets + context from date range")
    parser.add_argument(
        "--start_date",
        type=str,
        default="2019-05-22",
        help="start date in format YYYY-MM-DD",
    )
    parser.add_argument(
        "--extra_days",
        type=int,
        default=15,
        help="number of days to include from start date and after",
    )
    parser.add_argument(
        "--data_path",
        type=str,
        default="da_stopwords_part?_2019-01-01_2020-12-31.ndjson",
        help='path to ndjson file; for all: "*.ndjson"',
    )
    parser.add_argument(
        "--context_len",
        type=int,
        default=4,
        help="number of tweets to include in context",
    )
    args = parser.parse_args()

    path = os.path.join("/data", "004_twitter-stopword", args.data_path)

    date_range = range_of_dates(args.start_date, args.extra_days)
    print(date_range)
    print(
        f"Looking for tweets from {args.start_date} and {args.extra_days} days forward (to and indcluding: {datetime.datetime.strptime(args.start_date, '%Y-%m-%d').date() + datetime.timedelta(days=args.extra_days-1)})",
    )
    tweets = []

    for i, post in enumerate(ndjson_gen(path)):
        # get the date of the tweet and append the tweet if it is within the date range
        date_ = datetime.datetime.strptime(post["created_at"][0:10], "%Y-%m-%d").date()
        if date_ in date_range:
            tweets.append(post)

    print(f"Found {len(tweets)} tweets in the date range (out of {i} tweets total)")
    """#run the tweet thread extraction contexts =
    context_window_thread(tweets, args.context_len, args.start_date,
    args.extra_days)

    #write contests to ndjson
    with open(os.path.join(f"tweetcontexts_{datetime.datetime.strptime(args.start_date, '%Y-%m-%d').date() }_{datetime.datetime.strptime(args.start_date, '%Y-%m-%d').date() + datetime.timedelta(days=args.extra_days-1)}.ndjson"), "w") as f:
        ndjson.dump(contexts, f) #the ndjson of contexts has the format: List[Dict[str, Any]]
    """
    # CHATGPT
    # run the tweet thread extraction
    contexts = context_window_thread(tweets, args.context_len)

    # write contests to ndjson
    with open(
        os.path.join(
            f"tweet_threads_{datetime.datetime.strptime(args.start_date, '%Y-%m-%d').date() }_{datetime.datetime.strptime(args.start_date, '%Y-%m-%d').date() + datetime.timedelta(days=args.extra_days-1)}.ndjson",
        ),
        "w",
    ) as f:
        ndjson.dump(
            contexts,
            f,
        )  # the ndjson of contexts has the format: List[Dict[str, Any]]



import pickle

from ..model import BERTBiLSTM, Multi2OIE


def get_models(
    bert_config,
    pred_n_labels=3,
    arg_n_labels=9,
    n_arg_heads=8,
    n_arg_layers=4,
    lstm_dropout=0.3,
    mh_dropout=0.1,
    pred_clf_dropout=0.0,
    arg_clf_dropout=0.3,
    pos_emb_dim=64,
    use_lstm=False,
    device=None,
):
    if not use_lstm:
        return Multi2OIE(
            bert_config=bert_config,
            mh_dropout=mh_dropout,
            pred_clf_dropout=pred_clf_dropout,
            arg_clf_dropout=arg_clf_dropout,
            n_arg_heads=n_arg_heads,
            n_arg_layers=n_arg_layers,
            pos_emb_dim=pos_emb_dim,
            pred_n_labels=pred_n_labels,
            arg_n_labels=arg_n_labels,
        ).to(device)
    else:
        return BERTBiLSTM(
            bert_config=bert_config,
            lstm_dropout=lstm_dropout,
            pred_clf_dropout=pred_clf_dropout,
            arg_clf_dropout=arg_clf_dropout,
            pos_emb_dim=pos_emb_dim,
            pred_n_labels=pred_n_labels,
            arg_n_labels=arg_n_labels,
        ).to(device)


def load_pkl(path):
    with open(path, "rb") as f:
        return pickle.load(f)


def get_word2piece(sentence, tokenizer):
    words = sentence.split(" ")
    word2piece = {idx: list() for idx in range(len(words))}
    sentence_pieces = list()
    piece_idx = 1
    for word_idx, word in enumerate(words):
        pieces = tokenizer.tokenize(word)
        sentence_pieces += pieces
        for piece_idx_added, piece in enumerate(pieces):
            word2piece[word_idx].append(piece_idx + piece_idx_added)
        piece_idx += len(pieces)
    assert len(sentence_pieces) == piece_idx - 1
    return word2piece



from conspiracies.docprocessing.relationextraction.data_classes import (
    SpanTriplet,  # noqa F401
    StringTriplet,  # noqa F401
    DocTriplets,  # noqa F401
)
from .prompt_templates import (  # noqa F401
    MarkdownPromptTemplate1,
    MarkdownPromptTemplate2,
    PromptTemplate,
    PromptTemplate1,
    PromptTemplate2,
    XMLStylePromptTemplate,
    chatGPTPromptTemplate,
)

from .prompt_relation_component import (  # noqa F401
    create_prompt_relation_extraction_component,
    score_open_relations,
)



import json
import os

from conspiracies.common.fileutils import iter_lines_of_files
from conspiracies.corpusprocessing import umap_hdb
from conspiracies.docprocessing.docprocessor import DocProcessor
from conspiracies.document import Document
from conspiracies.pipeline.config import PipelineConfig
from conspiracies.preprocessing.csv import CsvPreprocessor
from conspiracies.preprocessing.infomedia import InfoMediaPreprocessor
from conspiracies.preprocessing.preprocessor import Preprocessor
from conspiracies.preprocessing.text import TextFilePreprocessor
from conspiracies.preprocessing.tweets import TweetsPreprocessor
from conspiracies.visualization.graph import create_network_graph, get_nodes_edges


class Pipeline:
    def __init__(self, config: PipelineConfig):
        self.project_name = config.base.project_name
        self.input_path = config.preprocessing.input_path
        self.config = config
        print("Initialized Pipeline with config:", config)
        self.output_path = os.path.join(self.config.base.output_root, self.project_name)
        os.makedirs(self.output_path, exist_ok=True)

    def run(self):
        if self.config.preprocessing.enabled:
            if self.input_path is None:
                raise ValueError("'input_path' must be provided for preprocessing!")
            self.preprocessing()

        if self.config.docprocessing.enabled:
            self.docprocessing(continue_from_last=True)

        if self.config.corpusprocessing.enabled:
            self.corpusprocessing()

    def _get_preprocessor(self) -> Preprocessor:
        config = self.config.preprocessing
        doc_type = config.doc_type.lower()
        if doc_type == "text":
            return TextFilePreprocessor(**config.extra)
        if doc_type == "csv":
            return CsvPreprocessor(
                metadata_fields=config.metadata_fields,
                **config.extra,
            )
        elif doc_type == "tweets":
            return TweetsPreprocessor(
                metadata_fields=config.metadata_fields,
                **config.extra,
            )
        elif doc_type == "infomedia":
            return InfoMediaPreprocessor(metadata_fields=config.metadata_fields)
        else:
            raise ValueError(
                f"Unknown document type: {doc_type}. "
                f"Available types are: tweets, infomedia",
            )

    def preprocessing(self) -> None:
        preprocessor = self._get_preprocessor()
        preprocessor.preprocess_docs(
            self.input_path,
            f"{self.output_path}/preprocessed.ndjson",
        )

    def _get_docprocessor(self) -> DocProcessor:
        return DocProcessor(
            triplet_extraction=self.config.docprocessing.triplet_extraction_method,
        )

    def docprocessing(self, continue_from_last=False):
        docprocessor = self._get_docprocessor()
        docprocessor.process_docs(
            (
                json.loads(line, object_hook=Document)
                for line in iter_lines_of_files(
                    f"{self.output_path}/preprocessed.ndjson",
                )
            ),
            f"{self.output_path}/annotations.ndjson",
            continue_from_last=continue_from_last,
        )

    def corpusprocessing(self):
        # TODO: this process is kind of dumb with all the writing and reading of
        #  files etc., but for now just make it work. It comes from individual scripts
        #  and a lot of the logic of data structures happen in those read/writes. Also,
        #  a lot of data that we might want to save is thrown away, e.g. clsutered
        #  entities, or calculated/fetched on the go, e.g. node weights for graphs.
        docs = (
            json.loads(line)
            for line in iter_lines_of_files(
                f"{self.output_path}/annotations.ndjson",
            )
        )
        with open(
            f"{self.output_path}/triplets.csv",
            "w+",
        ) as out:
            for doc in docs:
                for triplet in doc["semantic_triplets"]:
                    triplet_fields = [
                        triplet["semantic_triplets"][field]["text"]
                        for field in ("subject", "predicate", "object")
                    ]
                    print(", ".join(triplet_fields), file=out)
        umap_hdb.main(
            f"{self.output_path}/triplets.csv",
            "danskBERT",
            dim=40,
            save=f"{self.output_path}/nodes_edges.json",
        )
        nodes, edges = get_nodes_edges(
            f"{self.output_path}/",
            "nodes_edges.json",
        )
        graph = create_network_graph(
            nodes,
            edges,
            save=f"{self.output_path}/graph",
        )
        graph_data = {
            "nodes": [
                {"label": node[0], "data": node[1]} for node in graph.nodes.data()
            ],
            "edges": [
                {"from ": edge[0], "to": edge[1], "data": edge[2]}
                for edge in graph.edges.data()
            ],
        }
        with open(f"{self.output_path}/graph.json", "w+") as out:
            json.dump(graph_data, out)



import os
from typing import List

import numpy as np
import torch
from tqdm import tqdm

from .other import bio
from .multi2oie_utils import get_cached_tokenizer


def extract(args, model, loader, output_path):
    model.eval()
    os.makedirs(output_path, exist_ok=True)
    extraction_path = os.path.join(output_path, "extraction.txt")
    tokenizer = get_cached_tokenizer(args["bert_config"])
    f = open(extraction_path, "w")

    for step, batch in tqdm(enumerate(loader), desc="eval_steps", total=len(loader)):
        token_strs = [[word for word in sent] for sent in np.asarray(batch[-2]).T]
        sentences = batch[-1]
        token_ids, att_mask = map(lambda x: x.to(args.device), batch[:-2])

        with torch.no_grad():
            """We will iterate B(batch_size) times because there are more than
            one predicate in one batch. In feeding to argument extractor, # of
            predicates takes a role as batch size.

            pred_logit: (B, L, 3)
            pred_hidden: (B, L, D)
            pred_tags: (B, P, L) ~ list of tensors, where P is # of predicate in each
                batch
            """
            pred_logit, pred_hidden = model.extract_predicate(
                input_ids=token_ids,
                attention_mask=att_mask,
            )
            pred_tags = torch.argmax(pred_logit, 2)
            pred_tags = bio.filter_pred_tags(pred_tags, token_strs)
            pred_tags = bio.get_single_predicate_idxs(pred_tags)
            pred_probs = torch.nn.Softmax(2)(pred_logit)

            # iterate B times (one iteration means extraction for one sentence)
            for (
                cur_pred_tags,
                cur_pred_hidden,
                cur_att_mask,
                cur_token_id,
                cur_pred_probs,
                token_str,
                sentence,
            ) in zip(
                pred_tags,
                pred_hidden,
                att_mask,
                token_ids,
                pred_probs,
                token_strs,
                sentences,
            ):
                # generate temporary batch for this sentence and feed to argument module
                cur_pred_masks = bio.get_pred_mask(cur_pred_tags).to(args.device)
                n_predicates = cur_pred_masks.shape[0]
                if n_predicates == 0:
                    continue  # if there is no predicate, we cannot extract.
                cur_pred_hidden = torch.cat(
                    n_predicates * [cur_pred_hidden.unsqueeze(0)],
                )
                cur_token_id = torch.cat(n_predicates * [cur_token_id.unsqueeze(0)])
                cur_arg_logit = model.extract_argument(
                    input_ids=cur_token_id,
                    predicate_hidden=cur_pred_hidden,
                    predicate_mask=cur_pred_masks,
                )

                # filter and get argument tags with highest probability
                cur_arg_tags = torch.argmax(cur_arg_logit, 2)
                cur_arg_probs = torch.nn.Softmax(2)(cur_arg_logit)
                cur_arg_tags = bio.filter_arg_tags(
                    cur_arg_tags,
                    cur_pred_tags,
                    token_str,
                )

                # get string tuples and write results
                cur_extractions, cur_extraction_idxs = bio.get_tuple(
                    sentence,
                    cur_pred_tags,
                    cur_arg_tags,
                    tokenizer,
                )
                cur_confidences = bio.get_confidence_score(
                    cur_pred_probs,
                    cur_arg_probs,
                    cur_extraction_idxs,
                )
                for extraction, confidence in zip(cur_extractions, cur_confidences):
                    if args.binary:
                        f.write(
                            "\t".join([sentence] + [str(1.0)] + extraction[:3]) + "\n",
                        )
                    else:
                        f.write(
                            "\t".join([sentence] + [str(confidence)] + extraction)
                            + "\n",
                        )
        f.close()
    print("\nExtraction Done.\n")


def extract_to_list(args, model, loader):
    model.eval()
    tokenizer = get_cached_tokenizer(args["bert_config"])

    out = {}
    out["sentence"] = []
    out["confidence"] = []
    out["extraction"] = []
    out["extraction_3"] = []

    for step, batch in tqdm(enumerate(loader), desc="eval_steps", total=len(loader)):
        token_strs = [[word for word in sent] for sent in np.asarray(batch[-2]).T]
        sentences = batch[-1]
        token_ids, att_mask = map(lambda x: x.to(args["device"]), batch[:-2])

        with torch.no_grad():
            """We will iterate B(batch_size) times because there are more than
            one predicate in one batch. In feeding to argument extractor, # of
            predicates takes a role as batch size.

            pred_logit: (B, L, 3)
            pred_hidden: (B, L, D)
            pred_tags: (B, P, L) ~ list of tensors, where P is # of predicate in each
                batch
            """
            pred_logit, pred_hidden = model.extract_predicate(
                input_ids=token_ids,
                attention_mask=att_mask,
            )
            pred_tags = torch.argmax(pred_logit, 2)
            pred_tags = bio.filter_pred_tags(pred_tags, token_strs)
            pred_tags = bio.get_single_predicate_idxs(pred_tags)
            pred_probs = torch.nn.Softmax(2)(pred_logit)

            # iterate B times (one iteration means extraction for one sentence)
            for (
                cur_pred_tags,
                cur_pred_hidden,
                cur_att_mask,
                cur_token_id,
                cur_pred_probs,
                token_str,
                sentence,
            ) in zip(
                pred_tags,
                pred_hidden,
                att_mask,
                token_ids,
                pred_probs,
                token_strs,
                sentences,
            ):
                # generate temporary batch for this sentence and feed to argument module
                cur_pred_masks = bio.get_pred_mask(cur_pred_tags).to(args["device"])
                n_predicates = cur_pred_masks.shape[0]
                if n_predicates == 0:
                    continue  # if there is no predicate, we cannot extract.
                cur_pred_hidden = torch.cat(
                    n_predicates * [cur_pred_hidden.unsqueeze(0)],
                )
                cur_token_id = torch.cat(n_predicates * [cur_token_id.unsqueeze(0)])
                cur_arg_logit = model.extract_argument(
                    input_ids=cur_token_id,
                    predicate_hidden=cur_pred_hidden,
                    predicate_mask=cur_pred_masks,
                )

                # filter and get argument tags with highest probability
                cur_arg_tags = torch.argmax(cur_arg_logit, 2)
                cur_arg_probs = torch.nn.Softmax(2)(cur_arg_logit)
                cur_arg_tags = bio.filter_arg_tags(
                    cur_arg_tags,
                    cur_pred_tags,
                    token_str,
                )

                # get string tuples and write results
                cur_extractions, cur_extraction_idxs = bio.get_tuple(
                    sentence,
                    cur_pred_tags,
                    cur_arg_tags,
                    tokenizer,
                )
                cur_confidences = bio.get_confidence_score(
                    cur_pred_probs,
                    cur_arg_probs,
                    cur_extraction_idxs,
                )
                for extraction, confidence in zip(cur_extractions, cur_confidences):
                    out["sentence"].append(sentence)
                    out["confidence"].append(confidence)
                    out["extraction"].append(extraction)
                    out["extraction_3"].append(extraction[:3])

    print("\nExtraction Done.\n")
    return out


def simple_extract(model, loader, device):
    tokenizer = get_cached_tokenizer("bert-base-multilingual-cased")
    triplet = []
    conf = []
    for step, batch in tqdm(enumerate(loader), desc="eval_steps", total=len(loader)):
        token_strs = [[word for word in sent] for sent in np.asarray(batch[-2]).T]
        sentences = batch[-1]
        token_ids, att_mask = map(lambda x: x.to(device), batch[:-2])

        with torch.no_grad():
            """We will iterate B(batch_size) times because there are more than
            one predicate in one batch. In feeding to argument extractor, # of
            predicates takes a role as batch size.

            pred_logit: (B, L, 3)
            pred_hidden: (B, L, D)
            pred_tags: (B, P, L) ~ list of tensors, where P is # of predicate in each
                batch
            """
            pred_logit, pred_hidden = model.extract_predicate(
                input_ids=token_ids,
                attention_mask=att_mask,
            )
            pred_tags = torch.argmax(pred_logit, 2)
            pred_tags = bio.filter_pred_tags(pred_tags, token_strs)
            pred_tags = bio.get_single_predicate_idxs(pred_tags)
            pred_probs = torch.nn.Softmax(2)(pred_logit)

            # iterate B times (one iteration means extraction for one sentence)
            for (
                cur_pred_tags,
                cur_pred_hidden,
                cur_att_mask,
                cur_token_id,
                cur_pred_probs,
                token_str,
                sentence,
            ) in zip(
                pred_tags,
                pred_hidden,
                att_mask,
                token_ids,
                pred_probs,
                token_strs,
                sentences,
            ):
                # generate temporary batch for this sentence and feed to argument module
                cur_pred_masks = bio.get_pred_mask(cur_pred_tags).to(device)
                n_predicates = cur_pred_masks.shape[0]
                if n_predicates == 0:
                    continue  # if there is no predicate, we cannot extract.
                cur_pred_hidden = torch.cat(
                    n_predicates * [cur_pred_hidden.unsqueeze(0)],
                )
                cur_token_id = torch.cat(n_predicates * [cur_token_id.unsqueeze(0)])
                cur_arg_logit = model.extract_argument(
                    input_ids=cur_token_id,
                    predicate_hidden=cur_pred_hidden,
                    predicate_mask=cur_pred_masks,
                )

                # filter and get argument tags with highest probability
                cur_arg_tags = torch.argmax(cur_arg_logit, 2)
                cur_arg_probs = torch.nn.Softmax(2)(cur_arg_logit)
                cur_arg_tags = bio.filter_arg_tags(
                    cur_arg_tags,
                    cur_pred_tags,
                    token_str,
                )

                # get string tuples and write results
                cur_extractions, cur_extraction_idxs = bio.get_tuple(
                    sentence,
                    cur_pred_tags,
                    cur_arg_tags,
                    tokenizer,
                )
                cur_confidences = bio.get_confidence_score(
                    cur_pred_probs,
                    cur_arg_probs,
                    cur_extraction_idxs,
                )

                tmp_triplet = []
                tmp_confidence = []
                for extraction, confidence in zip(cur_extractions, cur_confidences):
                    tmp_triplet.append(extraction[:3])
                    tmp_confidence.append(confidence)

                triplet.append(tmp_triplet)
                conf.append(tmp_confidence)

    return (triplet, conf)


def extract_to_dict(args, model, loader):
    model.eval()
    tokenizer = get_cached_tokenizer(args["bert_config"])

    out = {}
    out["sentence"] = []
    out["wordpieces"] = []
    out["confidence"] = []
    out["extraction_span"] = []
    out["extraction"] = []

    for step, batch in enumerate(loader):
        token_strs = [[word for word in sent] for sent in np.asarray(batch[-2]).T]
        sentences = batch[-1]
        token_ids, att_mask = map(lambda x: x.to(args["device"]), batch[:-2])

        with torch.no_grad():
            """We will iterate B(batch_size) times because there are more than
            one predicate in one batch. In feeding to argument extractor, # of
            predicates takes a role as batch size.

            pred_logit: (B, L, 3)
            pred_hidden: (B, L, D)
            pred_tags: (B, P, L) ~ list of tensors, where P is # of predicate in each
                batch
            """
            pred_logit, pred_hidden = model.extract_predicate(
                input_ids=token_ids,
                attention_mask=att_mask,
            )
            pred_tags = torch.argmax(pred_logit, 2)
            pred_tags = bio.filter_pred_tags(pred_tags, token_strs)
            pred_tags = bio.get_single_predicate_idxs(pred_tags)
            pred_probs = torch.nn.Softmax(2)(pred_logit)

            # iterate B times (one iteration means extraction for one sentence)
            for (
                cur_pred_tags,
                cur_pred_hidden,
                cur_att_mask,
                cur_token_id,
                cur_pred_probs,
                token_str,
                sentence,
            ) in zip(
                pred_tags,
                pred_hidden,
                att_mask,
                token_ids,
                pred_probs,
                token_strs,
                sentences,
            ):
                # generate temporary batch for this sentence and feed to argument module
                cur_pred_masks = bio.get_pred_mask(cur_pred_tags).to(args["device"])
                n_predicates = cur_pred_masks.shape[0]
                if n_predicates == 0:
                    continue  # if there is no predicate, we cannot extract.
                cur_pred_hidden = torch.cat(
                    n_predicates * [cur_pred_hidden.unsqueeze(0)],
                )
                cur_token_id = torch.cat(n_predicates * [cur_token_id.unsqueeze(0)])
                cur_arg_logit = model.extract_argument(
                    input_ids=cur_token_id,
                    predicate_hidden=cur_pred_hidden,
                    predicate_mask=cur_pred_masks,
                )

                # filter and get argument tags with highest probability
                cur_arg_tags = torch.argmax(cur_arg_logit, 2)
                cur_arg_probs = torch.nn.Softmax(2)(cur_arg_logit)
                cur_arg_tags = bio.filter_arg_tags(
                    cur_arg_tags,
                    cur_pred_tags,
                    token_str,
                )

                # get string tuples and write results
                cur_extractions, cur_extraction_idxs = bio.get_tuple(
                    sentence,
                    cur_pred_tags,
                    cur_arg_tags,
                    tokenizer,
                )
                cur_confidences = bio.get_confidence_score(
                    cur_pred_probs,
                    cur_arg_probs,
                    cur_extraction_idxs,
                )

                def _remove_pad_tokens(wordpieces: List[str]):
                    return [token for token in wordpieces if token != "[PAD]"]

                def _flip_pos(t: tuple):
                    # swap start and end
                    t[0], t[1] = t[1], t[0]  # type: ignore
                    return t

                out["sentence"].append(sentence)
                out["wordpieces"].append(token_str)
                out["confidence"].append(cur_confidences)
                out["extraction_span"].append(
                    [_flip_pos(cur_idx[:3]) for cur_idx in cur_extraction_idxs],
                )
                out["extraction"].append(
                    [
                        _flip_pos(cur_extraction[:3])
                        for cur_extraction in cur_extractions
                    ],
                )

    return out
    return out



from typing import TypedDict, Optional


class Document(TypedDict):
    id: str
    metadata: dict
    text: str
    context: Optional[str]


CONTEXT_END_MARKER = "[CONTEXT_END]"


def text_with_context(doc: Document) -> str:
    if doc["context"] is None:
        return doc["text"]
    return "\n".join([doc["context"], CONTEXT_END_MARKER, doc["text"]])


def remove_context(text: str) -> str:
    return text.split(CONTEXT_END_MARKER)[-1].strip()



from pathlib import Path

import pytest
import spacy
from conspiracies import docs_from_jsonl, docs_to_jsonl
from conspiracies.docprocessing.relationextraction.gptprompting import (
    DocTriplets,
    SpanTriplet,
)
from spacy.tokens import Doc

from .utils import docs_with_triplets  # noqa: F401


@pytest.fixture()
def path():
    data_path = Path(__file__).parent / "test_data" / "triplets.jsonl"
    return data_path


@pytest.fixture()
def nlp():
    nlp = spacy.blank("en")
    nlp.add_pipe("sentencizer")
    return nlp


def test_docs_from_jsonl(path, nlp):
    docs = docs_from_jsonl(path, nlp=nlp)

    for doc in docs:
        assert isinstance(doc, Doc)
        triplets = doc._.relation_triplets
        for triplet in triplets:
            assert isinstance(triplet, SpanTriplet)
        for triplet in triplets:
            assert isinstance(triplet, SpanTriplet)


def test_docs_to_jsonl(nlp, docs_with_triplets):  # noqa: F811
    docs = docs_with_triplets
    docs_to_jsonl(docs, "test.jsonl", include_span_heads=False)
    _docs = docs_from_jsonl("test.jsonl", nlp=nlp)

    assert len(docs) == len(_docs)

    for doc, _doc in zip(docs, _docs):
        assert doc.text == _doc.text
        triplets = doc._.relation_triplets
        _triplets = _doc._.relation_triplets

        assert isinstance(triplets, DocTriplets)
        assert isinstance(_triplets, DocTriplets)

        assert len(triplets) == len(_triplets)
        for triplet, _triplet in zip(triplets, _triplets):
            assert triplet.is_string_match(_triplet)

    # clean up by removing the file
    Path("test.jsonl").unlink()



"""Pydantic data classes for the relation extraction using prompt-based
models."""

from copy import copy
from difflib import SequenceMatcher
from functools import partial
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import spacy
from pydantic import BaseModel, Extra
from spacy import displacy
from spacy.tokens import Doc, Span, Token


def get_text(span: Iterable[Token], ignore_spaces: bool, lowercase: bool) -> str:
    """Gets the text of a span or doc.

    Args:
        span (Iterable[Token]): a spacy span or doc.
        ignore_spaces (bool): whether to ignore spaces.
        lowercase (bool): whether to lowercase the text.

    Returns:
        str: the text of the span or doc.
    """
    if ignore_spaces:
        span_str = " ".join([t.text.strip() for t in span])
    else:
        span_str = "".join([t.text_with_ws for t in span])
    if lowercase:
        span_str = span_str.lower()
    return span_str


def subspan_of_span(
    subspan: Union[Span, Doc],
    span: Union[Span, Doc],
    lowercase: bool = False,
    ignore_spaces: bool = False,
) -> List[Span]:
    """Checks if a token is contained in a span. This function assumes that the
    token is not from the span and therefore.

    Args:
        subspan (Union[Span, Doc]): a spacy span or doc to check if it is contained
            wihtin span.
        span (Union[Span, Doc]): a spacy span (or Doc) to check if the subspan is
            contained within.

    Returns:
        List[Span]: a list of spans from span that are (string-)equal to the subspan.
    """
    if not span:
        return []

    _get_text = partial(get_text, ignore_spaces=ignore_spaces, lowercase=lowercase)
    if ignore_spaces:
        subspan = [t for t in subspan if t.text.strip() != ""]  # type: ignore
        span = [t for t in span if t.text.strip() != ""]  # type: ignore
    span_len = len(subspan)
    _spans = [span[i : i + span_len] for i in range(len(span) - span_len + 1)]

    subspan_text = _get_text(subspan)

    potential_spans = []
    for _span in _spans:
        _span_text = _get_text(_span)
        if subspan_text == _span_text:
            potential_spans.append(_span)

    if ignore_spaces:  # reconstruct to spans instead of list[token]
        try:
            doc = span[0].doc
            potential_spans = [
                doc[span[0].i : span[-1].i + 1] for span in potential_spans
            ]
        except IndexError:
            return []

    return potential_spans


class StringTriplet(BaseModel):
    class Config:
        extra = Extra.forbid

    subject: str
    predicate: str
    object: str
    subject_char_span: Optional[Tuple[int, int]] = None
    predicate_char_span: Optional[Tuple[int, int]] = None
    object_char_span: Optional[Tuple[int, int]] = None
    text: Optional[str] = None

    @property
    def triplet(self) -> Tuple[str, str, str]:
        return (self.subject, self.predicate, self.object)

    @property
    def char_spans(
        self,
    ) -> Tuple[
        Optional[Tuple[int, int]],
        Optional[Tuple[int, int]],
        Optional[Tuple[int, int]],
    ]:  # noqa: E501
        return (self.subject_char_span, self.predicate_char_span, self.object_char_span)

    def has_ranges(self) -> bool:
        return all(r is not None for r in self.char_spans)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, StringTriplet):
            if self.has_ranges() and other.has_ranges():
                has_same_char_span = self.char_spans == other.char_spans
            else:
                has_same_char_span = True

            return (
                self.triplet == other.triplet
                and self.text == other.text
                and has_same_char_span
            )
        return False


def _lcs_size(a: Sequence, b: Sequence) -> int:
    """Return the size of the longest common subsequence."""
    s = SequenceMatcher(None, a, b)
    return s.find_longest_match().size


class SpanTriplet(BaseModel):
    """A class for a semantic triplet with spans.

    Args:
        subject (Span): Subject of the triplet.
        predicate (Span): Predicate of the triplet.
        object (Span): Object of the triplet.
    """

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.forbid

    subject: Span
    predicate: Span
    object: Span

    @property
    def triplet(self) -> Tuple[Span, Span, Span]:
        """Returns the triplet as a tuple."""
        return self.subject, self.predicate, self.object

    @property
    def span(self) -> Span:
        """Returns the span of the triplet."""
        min_start = min([s.start for s in self.triplet])
        max_end = max([s.end for s in self.triplet])

        return self.subject.doc[min_start:max_end]

    @property
    def sentence(self) -> Span:
        """Returns the sentence of the triplet."""
        span = self.span
        doc = span.doc

        for sent in doc.sents:
            if sent.start <= span.start and sent.end >= span.end:
                return sent
            else:
                return doc[:]

    @property
    def doc(self) -> Doc:
        """Returns the doc of the triplet."""
        return self.subject.doc

    def __visualize_w_overlap(self):
        """Visualizes the triplet using displacy."""
        colors = {
            "SUBJECT": "#7aecec",
            "PREDICATE": "#ff9561",
            "OBJECT": "#aa9cfc",
        }
        spans = [
            {
                "start_token": s.start,
                "end_token": s.end,
                "label": s.label_,
            }
            for s in self.triplet
        ]
        ex = [
            {
                "text": self.doc.text,
                "spans": spans,
                "tokens": [t.text for t in self.doc],
            },
        ]
        html = displacy.render(
            ex,
            style="span",
            manual=True,
            options={"colors": colors},
        )
        return html

    def __visualize_no_overlap(self):
        """Visualizes the triplet using displacy."""
        self.subject.label_ = "SUBJECT"
        self.predicate.label_ = "PREDICATE"
        self.object.label_ = "OBJECT"
        colors = {
            "SUBJECT": "#7aecec",
            "PREDICATE": "#ff9561",
            "OBJECT": "#aa9cfc",
        }
        sent = self.sentence
        copy_doc = self.doc[:].as_doc()  # to avoid overwriting the original doc
        options = {"ents": ["SUBJECT", "PREDICATE", "OBJECT"], "colors": colors}
        copy_doc.ents = [self.subject, self.predicate, self.object]
        viz_doc = copy_doc[sent.start : sent.end].as_doc()
        html = displacy.render(viz_doc, style="ent", options=options)
        return html

    def is_string_match(self, other) -> bool:
        if not isinstance(other, SpanTriplet):
            return False

        triplet_is_string_match = all(
            s1.text == s2.text for s1, s2 in zip(self.triplet, other.triplet)
        )
        return triplet_is_string_match

    def visualize(self, overlap: Optional[bool] = None):
        """Visualizes the triplet using displacy."""
        if overlap is None:
            try:
                return self.__visualize_no_overlap()
            except ValueError:
                return self.__visualize_w_overlap()
        if overlap:
            return self.__visualize_w_overlap()
        else:
            return self.__visualize_no_overlap()

    @staticmethod
    def span_to_json(span: Union[Span, Doc]) -> Dict[str, Any]:
        if isinstance(span, Doc):
            span = span[:]
        return {
            "text": span.text,
            "start": span.start,
            "end": span.end,
        }

    @staticmethod
    def span_from_json(json: dict, doc: Doc) -> Span:
        span = doc[json["start"] : json["end"]]
        return span

    def to_dict(self, include_doc=True, include_span_heads=False) -> Dict[str, Any]:
        if include_doc:
            data = self.span.doc.to_json()
        else:
            data = {}
        data["semantic_triplets"] = self.dict()
        data["semantic_triplets"]["subject"] = self.span_to_json(self.subject)
        data["semantic_triplets"]["predicate"] = self.span_to_json(self.predicate)
        data["semantic_triplets"]["object"] = self.span_to_json(self.object)
        if include_span_heads and Span.has_extension("most_common_ancestor"):
            subject_head = self.subject._.most_common_ancestor.text
            data["semantic_triplets"]["subject"]["head"] = subject_head
            object_head = self.object._.most_common_ancestor.text
            data["semantic_triplets"]["object"]["head"] = object_head

        return data

    @staticmethod
    def from_tuple(
        triplet: Tuple[Span, Span, Span],
    ):
        subject = triplet[0]
        predicate = triplet[1]
        object = triplet[2]
        subject.label_ = "SUBJECT"
        predicate.label_ = "PREDICATE"
        object.label_ = "OBJECT"

        return SpanTriplet(
            subject=subject,
            predicate=predicate,
            object=object,
        )

    @staticmethod
    def from_dict(
        data: Dict[str, Any],
        nlp: spacy.Language,
        doc: Optional[Doc] = None,
    ):
        if doc is None:
            doc = Doc(nlp.vocab).from_json(data)  # type: ignore
        subject = SpanTriplet.span_from_json(data["semantic_triplets"]["subject"], doc)
        predicate = SpanTriplet.span_from_json(
            data["semantic_triplets"]["predicate"],
            doc,
        )
        object = SpanTriplet.span_from_json(data["semantic_triplets"]["object"], doc)
        subject.label_ = "SUBJECT"
        predicate.label_ = "PREDICATE"
        object.label_ = "OBJECT"

        return SpanTriplet(
            subject=subject,
            predicate=predicate,
            object=object,
        )

    @staticmethod
    def span_triplet_from_text_triplet(
        triplet: Tuple[str, str, str],
        span: Union[Span, Doc],
        lowercase: Optional[bool] = None,
    ) -> Optional["SpanTriplet"]:
        """Checks if the triplet contained within the doc based on text
        overlap.

        Args:
            triplet (Tuple[str, str, str]): A semantic triplet to check if contained
                within the span. A triplet contains subject, predicate, and object (in
                that order).
            span (Union[Span, Doc]): A spacy span or doc.
            lowercase (Optional[bool]): Whether to use lowercase
                comparison. Defaults to None. If None it will first try to match the
                triplet in the original case and then fallback to lowercase.

        Returns:
            Optional[SpanTriplet]: A SpanTriplet object. Returns None if the triplet is
                not contained in the span.
        """
        if any(t.strip() == "" for t in triplet):
            return None

        if lowercase is None:
            spantriplet = SpanTriplet.span_triplet_from_text_triplet(
                triplet,
                span,
                lowercase=False,
            )
            if spantriplet is not None:
                return spantriplet
            return SpanTriplet.span_triplet_from_text_triplet(
                triplet,
                span,
                lowercase=True,
            )

        if lowercase:
            text = span.text.lower()
            subj, pred, obj = [t.lower() for t in triplet]
        else:
            text = span.text
            subj, pred, obj = triplet
        subj_start = text.find(subj)
        if subj_start == -1:
            return None
        subj_end = subj_start + len(subj)

        # first try to find the object after the subject
        obj_start = text[subj_end:].find(obj)
        if obj_start != -1:
            obj_start = text.find(obj)
        if obj_start == -1:
            return None
        obj_end = obj_start + len(obj)

        # first try to find the predicate in between the subject and object
        start_pred = text[subj_end:obj_end].find(pred)
        if start_pred != -1:
            start_pred = text.find(pred)
        if start_pred == -1:
            return None
        end_pred = start_pred + len(pred)

        # create spans from ranges
        return SpanTriplet.from_char_span(
            span,
            (subj_start, subj_end),
            (start_pred, end_pred),
            (obj_start, obj_end),
        )

    @staticmethod
    def from_char_span(
        doc: Union[Span, Doc],
        subj_char_span: Tuple[int, int],
        pred_char_span: Tuple[int, int],
        obj_char_span: Tuple[int, int],
    ) -> Optional["SpanTriplet"]:
        """Creates a SpanTriplet from a StringTriplet.

        Args:
            doc: A spacy span or doc.
            str_triplet: A StringTriplet object.

        Returns:
            A SpanTriplet object. Returns None if the triplet is not contained in the
                span.
        """
        triplet = []
        for label, (start, end) in zip(
            ["SUBJECT", "PREDICATE", "OBJECT"],
            [subj_char_span, pred_char_span, obj_char_span],
        ):
            span = doc.char_span(start, end, label=label)  # type: ignore
            if span is None:
                return None
            triplet.append(span)

        return SpanTriplet.from_tuple(triplet)  # type: ignore

    @staticmethod
    def span_triplet_from_span_triplet(
        triplet: Tuple[Span, Span, Span],
        span: Union[Span, Doc],
        lowercase: Optional[bool] = None,
        ignore_spaces: Optional[bool] = True,
    ) -> Optional["SpanTriplet"]:
        """Checks if the triplet contained within the doc based on overlap of
        span tokens.

        Args:
            triplet: A semantic triplet to check if contained
                within the span. A triplet contains subject, predicate, and object (in
                that order).
            span: A spacy span or doc.
            lowercase: Whether to use lowercase comparison. Defaults to
                None. If None it will first try to match the triplet in the original
                case and then fallback to lowercase.
            ignore_spaces: Whether to ignore spaces when comparing
                spans. Defaults to True. If None it will first try to match the triplet
                with spaces and then fallback to ignoring spaces.

        Returns:
            Optional[SpanTriplet]: A SpanTriplet object. Returns None if the triplet is
                not contained in the span.
        """

        if isinstance(span, Doc):
            span = span[:]
        if ignore_spaces is None:
            spantriplet = SpanTriplet.span_triplet_from_span_triplet(
                triplet,
                span,
                lowercase=lowercase,
                ignore_spaces=False,
            )
            if spantriplet is not None:
                return spantriplet
            return SpanTriplet.span_triplet_from_span_triplet(
                triplet,
                span,
                lowercase=lowercase,
                ignore_spaces=True,
            )
        if lowercase is None:
            spantriplet = SpanTriplet.span_triplet_from_span_triplet(
                triplet,
                span,
                lowercase=False,
            )
            if spantriplet is not None:
                return spantriplet
            return SpanTriplet.span_triplet_from_span_triplet(
                triplet,
                span,
                lowercase=True,
            )

        _subspan_of_span = partial(
            subspan_of_span,
            lowercase=lowercase,
            ignore_spaces=ignore_spaces,
        )

        subj, pred, obj = triplet
        cand_subj_span = _subspan_of_span(subj, span)
        if not cand_subj_span:
            return None
        subj_span = cand_subj_span[0]  # assume first match is correct

        # first assume object is after subject
        cand_obj_span = _subspan_of_span(obj, span.doc[subj_span.end : span.end])
        if not cand_obj_span:
            cand_obj_span = _subspan_of_span(obj, span)
        if not cand_obj_span:
            return None
        obj_span = cand_obj_span[0]  # assume first match is correct

        # first assume predicate is in between subject and object
        min_idx = min(subj_span.end, obj_span.end)
        max_idx = max(subj_span.start, obj_span.start)
        cand_pred_span = _subspan_of_span(pred, span.doc[min_idx:max_idx])
        if not cand_pred_span:  # fallback to entire span
            cand_pred_span = _subspan_of_span(pred, span)
            if not cand_pred_span:
                return None

            # try to find the predicate closest to the subject object span
            cand_pred_span_before = [
                span for span in cand_pred_span if span.end < min_idx
            ]
            if cand_pred_span_before:
                pred_span = sorted(cand_pred_span_before, key=lambda x: x.end)[-1]
            else:
                # otherwise take the first match
                pred_span = cand_pred_span[0]
        else:
            # assume the one closest to the object
            pred_span = sorted(
                cand_pred_span,
                key=lambda x: abs(x.start - obj_span.start),
            )[0]

        # update the subject span to be the closest to the predicate/obj
        min_idx = min(pred_span.end, obj_span.end)
        cand_subj_span = [s for s in cand_subj_span if s.end <= min_idx]
        if cand_subj_span:
            subj_span = sorted(cand_subj_span, key=lambda x: x.start)[-1]

        # label spans
        subj_span.label_ = "SUBJECT"
        pred_span.label_ = "PREDICATE"
        obj_span.label_ = "OBJECT"

        return SpanTriplet(
            subject=subj_span,
            predicate=pred_span,
            object=obj_span,
        )

    @staticmethod
    def from_doc(
        triplet: Union[Tuple[str, str, str], "StringTriplet"],
        doc: Union[Doc, Span],
        nlp: Optional[spacy.Language] = None,
        method: Optional[Literal["span", "text"]] = None,
        lowercase: Optional[bool] = None,
        ignore_spaces: Optional[bool] = True,
    ) -> Optional["SpanTriplet"]:
        """Converts the StringTriplet to a SpanTriplet. First checks if the
        span is contained within a singular sentence, then it checks if the
        span is contained within the entire document.

        Args:
            doc: Document of the text.
            triplet: A semantic triplet to check if contained within the span. A triplet
                contains subject, predicate, and object (in that order).
            nlp: Spacy language model. Defaults to None. In which case it will create a
                blank spacy model based on the language of the doc.
            method: Whether to use the span or text comparison to find valid span
                triplets. Defaults to None. If None it will first try to match the
                triplet based on the span and then fallback to text comparison.
            lowercase: Whether to lowercase the text before comparison. Defaults to
                None. If None it will first try to match the triplet in the original
                case and then fallback to lowercase.
            ignore_spaces: Whether to ignore spaces when comparing spans. Defaults to
                True. If None it will first try to match the triplet with spaces and
                then fallback to ignoring spaces.

        Returns:
            A SpanTriplet object. Returns None if the triplet is not contained in the
                span.
        """
        if isinstance(triplet, StringTriplet):
            if triplet.has_ranges():
                span_triplet = SpanTriplet.from_char_span(
                    doc,
                    triplet.subject_char_span,  # type: ignore
                    triplet.predicate_char_span,  # type: ignore
                    triplet.object_char_span,  # type: ignore
                )
                if span_triplet is not None:
                    return span_triplet
            triplet = triplet.triplet

        subject, predicate, object = triplet
        if any(t.strip() == "" for t in triplet):
            return None
        if method is None:
            span_triplet = SpanTriplet.from_doc(
                triplet,
                doc,
                nlp=nlp,
                method="span",
                lowercase=lowercase,
                ignore_spaces=ignore_spaces,
            )
            if span_triplet is not None:
                return span_triplet
            return SpanTriplet.from_doc(
                triplet,
                doc,
                nlp=nlp,
                method="text",
                lowercase=lowercase,
                ignore_spaces=ignore_spaces,
            )
        if nlp is None:
            if isinstance(doc, Span):
                lang = doc.doc.lang_
            else:
                lang = doc.lang_
            nlp = spacy.blank(lang)
            nlp.add_pipe("sentencizer")

        if method == "span":
            triplet = tuple([doc[:] for doc in nlp.pipe(triplet)])  # type: ignore
        else:
            triplet = triplet

        span_triplet_from_mapping = {
            "span": partial(
                SpanTriplet.span_triplet_from_span_triplet,
                ignore_spaces=ignore_spaces,
            ),
            "text": SpanTriplet.span_triplet_from_text_triplet,
        }

        span_triplet_from = span_triplet_from_mapping[method]

        for sent in doc.sents:  # type: ignore
            span_triplet = span_triplet_from(
                triplet,
                sent,
                lowercase=lowercase,  # type: ignore
            )
            if span_triplet is not None:
                return span_triplet
        return span_triplet_from(triplet, doc, lowercase=lowercase)  # type: ignore

    def normalized_span_overlap(self, other: "SpanTriplet") -> float:
        """Calculates the normalized span overlap between two span triplets.
        Note this assumes the self SpanTriplet is the reference.

        Args:
            other: The other span triplet.

        Returns:
            The normalized span overlap between the two span triplets. The normalized
                is normalized between 0 and 1.
        """
        _normalized_overlap = 0.0
        for s_t, o_t in zip(self.triplet, other.triplet):
            overlap = _lcs_size(s_t, o_t)  # type: ignore
            _normalized_overlap += overlap / len(s_t)
        return _normalized_overlap / 3

    def normalized_string_overlap(self, other: "SpanTriplet") -> float:
        """Calculates the normalized span overlap between two span triplets.
        Note this assumes the self SpanTriplet is the reference.

        Args:
            other: The other span triplet.

        Returns:
            The normalized span overlap between the two span triplets. The normalized
                is normalized between 0 and 1.
        """
        _normalized_overlap = 0.0
        for s_t, o_t in zip(self.triplet, other.triplet):
            overlap = _lcs_size(s_t.text, o_t.text)
            _normalized_overlap += overlap / len(s_t.text)
        return _normalized_overlap / 3

    def __eq__(self, other) -> bool:
        if not isinstance(other, SpanTriplet):
            return False

        # It does not make sense to compare Spans in different texts, but they can
        # be different Doc objects!
        if self.doc.text != other.doc.text:
            return False

        # Equality is checked for start, end and text such that two triplets in
        # different Doc objects can be considered equal
        triplet_is_equal = all(
            s1.start == s2.start and s1.end == s2.end and s1.text == s2.text
            for s1, s2 in zip(self.triplet, other.triplet)
        )
        return triplet_is_equal


class DocTriplets(BaseModel):
    """A class containing the relations of a documents."""

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.forbid

    span_triplets: List[SpanTriplet]
    doc: Doc

    @staticmethod
    def from_str_triplets(
        doc: Union[Doc, Span],
        triplets: List[StringTriplet],
    ) -> "DocTriplets":
        span_triplets = []
        for triplet in triplets:
            span_triplet = SpanTriplet.from_doc(triplet=triplet, doc=doc)
            if span_triplet is not None:
                span_triplets.append(span_triplet)
        return DocTriplets(span_triplets=span_triplets, doc=doc)

    def score_relations(self, reference: "DocTriplets") -> Dict[str, Any]:
        """Score the relations of the doctriplet against the relations of the
        current doctriplet."""
        self_relations = list(self)
        reference_relations = list(reference)
        self_relations.sort(key=lambda x: x.subject.text)
        reference_relations.sort(key=lambda x: x.subject.text)

        score = {
            "exact_span_match": 0,
            "exact_string_match": 0,
            "normalized_span_overlap": 0.0,
            "normalized_string_overlap": 0.0,
            "length_self": len(self_relations),
            "length_reference": len(reference_relations),
        }

        if score["length_reference"] == 0:
            return score

        _self_relations = copy(self_relations)
        missing_matches = {}

        for ref_i, ref_rel in enumerate(reference_relations):
            span_norm_overlap = []
            string_norm_overlap = []
            span_match_found = False
            string_match_found = False

            if not _self_relations:
                break

            for self_i, self_rel in enumerate(_self_relations):
                if self_rel == ref_rel:
                    span_match_found = True
                    break

                if self_rel.is_string_match(ref_rel):
                    string_match_found = True
                    s_norm_overlap = ref_rel.normalized_span_overlap(self_rel)

                if not string_match_found:
                    str_norm_overlap = ref_rel.normalized_string_overlap(self_rel)
                    s_norm_overlap = ref_rel.normalized_span_overlap(self_rel)
                    string_norm_overlap.append(str_norm_overlap)
                    span_norm_overlap.append(s_norm_overlap)

            if span_match_found:
                score["exact_span_match"] += 1
                score["exact_string_match"] += 1
                score["normalized_string_overlap"] += 1
                score["normalized_span_overlap"] += 1
                _self_relations.pop(self_i)
                continue
            if string_match_found:
                score["exact_string_match"] += 1
                score["normalized_string_overlap"] += 1
                score["normalized_span_overlap"] += s_norm_overlap
                _self_relations.pop(self_i)
                continue
            # if not match then just record the best match
            which_max = np.argmax(string_norm_overlap)
            missing_matches[ref_i] = {
                "span_norm_overlap": span_norm_overlap[which_max],
                "string_norm_overlap": string_norm_overlap[which_max],
                "self_i": which_max,
            }

        if missing_matches and _self_relations:
            # if there are missing matches and there are still self relations
            # then we can try to match them

            # sort the missing matches by the string norm overlap
            _missing_matches = [
                (k, v)
                for k, v in sorted(
                    missing_matches.items(),
                    key=lambda item: item[1]["string_norm_overlap"],  # type: ignore
                )
            ]

            while _missing_matches and _self_relations:
                ref_i, match = _missing_matches.pop()
                score["normalized_string_overlap"] += match["string_norm_overlap"]
                score["normalized_span_overlap"] += match["span_norm_overlap"]
                _self_relations.pop()  # just pop any self relation
        return score

    def __getitem__(self, index: int) -> SpanTriplet:
        return self.span_triplets[index]

    def __iter__(self) -> Iterator[SpanTriplet]:  # type: ignore
        return iter(self.span_triplets)

    def __len__(self) -> int:
        return len(self.span_triplets)

    def __add__(self, other: Union["DocTriplets", "SpanTriplet"]) -> "DocTriplets":
        if self.doc != other.doc:
            raise ValueError("Can only add DocTriplets from the same document.")
        if isinstance(other, SpanTriplet):
            self.span_triplets.append(other)
        else:
            self.span_triplets.extend(other.span_triplets)
        return self

    def __eq__(self, other: Any) -> bool:
        """Check if two DocTriplets are equal based on if all their span
        triplets are the same."""
        if not isinstance(other, DocTriplets):
            return False
        for s, o in zip(self, other):
            if s != o:
                return False
        return True


def span_to_idx(span: Span) -> Tuple[int, int]:
    return span.start, span.end


def idx_to_span(idx: Tuple[int, int], doc: Doc) -> Span:
    start, end = idx
    return doc[start:end]


def install_extensions(force=False) -> None:
    """Sets extensions on the SpaCy Doc class.

    Relation triplets are stored internally as index tuples, but they
    are created with getters and setters that map the index tuples to
    and from SpaCy Span objects. Heads, relations and tails are
    retrieved from the triplets. Confidence numbers are stored as is.
    """
    extensions = [
        "relation_triplet_idxs",
        "relation_triplets",
        "relation_head",
        "relation_relation",
        "relation_tail",
        "relation_confidence",
    ]
    if not force and all(Doc.has_extension(ext) for ext in extensions):
        return  # job's done!

    Doc.set_extension("relation_triplet_idxs", default=[], force=force)
    Doc.set_extension(
        "relation_triplets",
        setter=lambda doc, triplets: setattr(
            doc._,
            "relation_triplet_idxs",
            [
                tuple(span_to_idx(span) for span in span_triplet.triplet)
                for span_triplet in triplets
            ],
        ),
        getter=lambda doc: DocTriplets(
            span_triplets=[
                SpanTriplet.from_tuple(
                    (
                        idx_to_span(idx[0], doc),
                        idx_to_span(idx[1], doc),
                        idx_to_span(idx[2], doc),
                    ),
                )
                for idx in doc._.relation_triplet_idxs
            ],
            doc=doc,
        ),
        force=force,
    )
    Doc.set_extension(
        "relation_head",
        getter=lambda doc: [t.subject for t in doc._.relation_triplets],
        force=force,
    )
    Doc.set_extension(
        "relation_relation",
        getter=lambda doc: [t.predicate for t in doc._.relation_triplets],
        force=force,
    )
    Doc.set_extension(
        "relation_tail",
        getter=lambda doc: [t.object for t in doc._.relation_triplets],
        force=force,
    )
    Doc.set_extension("relation_confidence", default=None, force=force)



from typing import Iterator

import pytest

from conspiracies.document import Document
from conspiracies.preprocessing.preprocessor import Preprocessor


@pytest.fixture
def docs() -> Iterator[Document]:
    # if using yield, the pytest fixture does not work quite as expected
    docs = []
    for i in range(2):
        docs.append(
            Document(
                id=str(i),
                text=f"text {i}",
                metadata={
                    "one": "something",
                    "two": "something else",
                },
                context=None,
            ),
        )
    return (d for d in docs)


def test_metadata_filtering_removes_field(docs):
    preprocessor = Preprocessor(metadata_fields={"one"})
    filtered = list(preprocessor._filter_metadata(docs))
    assert len(filtered) == 2
    for doc in filtered:
        assert "one" in doc["metadata"] and "two" not in doc["metadata"]


def test_metadata_filtering_retains_all(docs):
    preprocessor = Preprocessor(metadata_fields={"*"})
    filtered = list(preprocessor._filter_metadata(docs))
    assert len(filtered) == 2
    for doc in filtered:
        assert all(key in doc["metadata"] for key in ("one", "two"))



import glob
import logging


def iter_lines_of_files(glob_pattern: str):
    files = glob.glob(glob_pattern, recursive=True)
    logging.debug(
        "The glob pattern '%s' resulted in the following files: %s",
        glob_pattern,
        files,
    )
    for file in files:
        with open(file) as f:
            for line in f:
                yield line



"""Functions for concatenating and splitting tweets and their context tweets
for coreference resolution."""

from typing import List
import ndjson
from glob import glob


def ndjson_gen(filepath: str):
    """Creates a generator for ndjson files and yields the tweets."""
    for file in glob(filepath):
        with open(file) as f:
            reader = ndjson.reader(f)
            for post in reader:
                yield post


def concat_context(context: List[dict], end_token: str = " [END] ") -> str:
    tweets_list = [tweet["text"] for tweet in context]
    return end_token.join(tweets_list)


def tweet_from_context(context: List[dict]) -> dict:
    return context[-1]


def tweet_from_context_text(context: str, end_token: str = " [END] ") -> str:
    tweets_list = context.split(end_token)
    return tweets_list[-1]


if __name__ == "__main__":
    # test functions
    context_path = "tweet_threads_2019-05-22_2019-06-05.ndjson"
    for i, context in enumerate(ndjson_gen(context_path)):
        if i == 5:
            break
        print("------ \n")
        concat_context_ = concat_context(context)
        print(concat_context_, "\n")
        tweet = tweet_from_context(context)
        tweet_text = tweet_from_context_text(concat_context_)
        print(tweet["text"] == tweet_text, "\n\n")



import glob
import os.path

from conspiracies.document import Document
from conspiracies.preprocessing.preprocessor import Preprocessor


class TextFilePreprocessor(Preprocessor):

    def __init__(self, basename_as_id=True):
        super().__init__()
        self.basename_as_id = basename_as_id

    def _do_preprocess_docs(self, glob_pattern: str):
        files = glob.glob(glob_pattern, recursive=True)
        for file in files:
            id_ = os.path.basename(file) if self.basename_as_id else file
            with open(file) as in_file:
                text = in_file.read()
            yield Document(
                id=id_,
                text=text,
                metadata={},
                context=None,
            )



import numpy as np
import torch

from . import utils

pred_tag2idx = {"P-B": 0, "P-I": 1, "O": 2}
arg_tag2idx = {
    "A0-B": 0,
    "A0-I": 1,
    "A1-B": 2,
    "A1-I": 3,
    "A2-B": 4,
    "A2-I": 5,
    "A3-B": 6,
    "A3-I": 7,
    "O": 8,
}


def get_pred_idxs(pred_tags):
    idxs = list()
    for pred_tag in pred_tags:
        idxs.append([idx.item() for idx in (pred_tag != 2).nonzero()])
    return idxs


def get_pred_mask(tensor):
    """Generate predicate masks by converting predicate index with 'O' tag to
    1. Other indexes are converted to 0 which means non-masking.

    :param tensor: predicate tagged tensor with the shape of (B, L),
        where B is the batch size, L is the sequence length.
    :return: masked binary tensor with the same shape.
    """
    res = tensor.bool()
    res[tensor == pred_tag2idx["O"]] = True
    res[tensor != pred_tag2idx["O"]] = False
    return res


def filter_pred_tags(pred_tags, tokens):
    """Filter useless tokens by converting them into 'Outside' tag. We treat
    'Inside' tag before 'Beginning' tag as meaningful signal, so changed them
    to 'Beginning' tag unlike [Stanovsky et al., 2018].

    :param pred_tags: predicate tags with the shape of (B, L).
    :param tokens: list format sentence pieces with the shape of (B, L)
    :return: tensor of filtered predicate tags with the same shape.
    """
    assert len(pred_tags) == len(tokens)
    assert len(pred_tags[0]) == len(tokens[0])

    # filter by tokens ([CLS], [SEP], [PAD] tokens should be allocated as 'O')
    for pred_idx, cur_tokens in enumerate(tokens):
        for tag_idx, token in enumerate(cur_tokens):
            if token in ["[CLS]", "[SEP]", "[PAD]"]:
                pred_tags[pred_idx][tag_idx] = pred_tag2idx["O"]

    # filter by tags
    pred_copied = pred_tags.clone()
    for pred_idx, cur_pred_tag in enumerate(pred_copied):
        flag = False
        tag_copied = cur_pred_tag.clone()
        for tag_idx, tag in enumerate(tag_copied):
            if not flag and tag == pred_tag2idx["P-B"]:
                flag = True
            elif not flag and tag == pred_tag2idx["P-I"]:
                pred_tags[pred_idx][tag_idx] = pred_tag2idx["P-B"]
                flag = True
            elif flag and tag == pred_tag2idx["O"]:
                flag = False
    return pred_tags


def filter_arg_tags(arg_tags, pred_tags, tokens):
    """Same as the description of @filter_pred_tags().

    :param arg_tags: argument tags with the shape of (B, L).
    :param pred_tags: predicate tags with the same shape. It is used to
        force predicate position to be allocated the 'Outside' tag.
    :param tokens: list of string tokens with the length of L. It is
        used to force special tokens like [CLS] to be allocated the
        'Outside' tag.
    :return: tensor of filtered argument tags with the same shape.
    """
    # filter by tokens ([CLS], [SEP], [PAD] tokens should be allocated as 'O')
    for arg_idx, cur_arg_tag in enumerate(arg_tags):
        for tag_idx, token in enumerate(tokens):
            if token in ["[CLS]", "[SEP]", "[PAD]"]:
                arg_tags[arg_idx][tag_idx] = arg_tag2idx["O"]

    # filter by tags
    arg_copied = arg_tags.clone()
    for arg_idx, (cur_arg_tag, cur_pred_tag) in enumerate(zip(arg_copied, pred_tags)):
        pred_idxs = [
            idx[0].item() for idx in (cur_pred_tag != pred_tag2idx["O"]).nonzero()
        ]
        arg_tags[arg_idx][pred_idxs] = arg_tag2idx["O"]
        cur_arg_copied = arg_tags[arg_idx].clone()
        flag_idx = 999
        for tag_idx, tag in enumerate(cur_arg_copied):
            if tag == arg_tag2idx["O"]:
                flag_idx = 999
                continue
            arg_n = torch.div(tag, 2, rounding_mode="floor")  # 0: A0 / 1: A1 / ...
            inside = tag % 2  # 0: begin / 1: inside
            if not inside and flag_idx != arg_n:
                flag_idx = arg_n
            # connect_args
            elif not inside and flag_idx == arg_n:
                arg_tags[arg_idx][tag_idx] = arg_tag2idx[f"A{arg_n}-I"]
            elif inside and flag_idx != arg_n:
                arg_tags[arg_idx][tag_idx] = arg_tag2idx[f"A{arg_n}-B"]
                flag_idx = arg_n
    return arg_tags


def get_max_prob_args(arg_tags, arg_probs):
    """Among predicted argument tags, remain only arguments with highest probs.
    The comparison of probability is made only between the same argument
    labels.

    :param arg_tags: argument tags with the shape of (B, L).
    :param arg_probs: argument softmax probabilities with the shape of
        (B, L, T), where B is the batch size, L is the sequence length,
        and T is the # of tag labels.
    :return: tensor of filtered argument tags with the same shape.
    """
    for cur_arg_tag, cur_probs in zip(arg_tags, arg_probs):
        cur_tag_probs = [cur_probs[idx][tag] for idx, tag in enumerate(cur_arg_tag)]
        for arg_n in range(4):
            b_tag = arg_tag2idx[f"A{arg_n}-B"]
            i_tag = arg_tag2idx[f"A{arg_n}-I"]
            flag = False
            total_tags = []
            cur_tags = []
            for idx, tag in enumerate(cur_arg_tag):
                if not flag and tag == b_tag:
                    flag = True
                    cur_tags.append(idx)
                elif flag and tag == i_tag:
                    cur_tags.append(idx)
                elif flag and tag == b_tag:
                    total_tags.append(cur_tags)
                    cur_tags = [idx]
                elif tag != b_tag or tag != i_tag:
                    total_tags.append(cur_tags)
                    cur_tags = []
                    flag = False
            max_idxs, max_prob = None, 0.0
            for idxs in total_tags:
                all_probs = [cur_tag_probs[idx].item() for idx in idxs]
                if len(all_probs) == 0:
                    continue
                cur_prob = all_probs[0]
                if cur_prob > max_prob:
                    max_prob = cur_prob
                    max_idxs = idxs
            if max_idxs is None:
                continue
            del_idxs = [
                idx
                for idx, tag in enumerate(cur_arg_tag)
                if (tag in [b_tag, i_tag]) and (idx not in max_idxs)
            ]
            cur_arg_tag[del_idxs] = arg_tag2idx["O"]
    return arg_tags


def get_single_predicate_idxs(pred_tags):
    """Divide each single batch based on predicted predicates. It is necessary
    for predicting argument tags with specific predicate.

    :param pred_tags: tensor of predicate tags with the shape of (B, L)
        EX >>> tensor([[2, 0, 0, 1, 0, 1, 0, 2, 2, 2],
                       [2, 2, 2, 0, 1, 0, 1, 2, 2, 2],
                       [2, 2, 2, 2, 2, 2, 2, 2, 0, 1]])

    :return: list of tensors with the shape of (B, P, L)
        the number P can be different for each batch.
        EX >>> [tensor([[2., 0., 2., 2., 2., 2., 2., 2., 2., 2.],
                        [2., 2., 0., 1., 2., 2., 2., 2., 2., 2.],
                        [2., 2., 2., 2., 0., 1., 2., 2., 2., 2.],
                        [2., 2., 2., 2., 2., 2., 0., 2., 2., 2.]]),
                tensor([[2., 2., 2., 0., 1., 2., 2., 2., 2., 2.],
                        [2., 2., 2., 2., 2., 0., 1., 2., 2., 2.]]),
                tensor([[2., 2., 2., 2., 2., 2., 2., 2., 0., 1.]])]
    """
    total_pred_tags = []
    for cur_pred_tag in pred_tags:
        cur_sent_preds = []
        begin_idxs = [
            idx[0].item() for idx in (cur_pred_tag == pred_tag2idx["P-B"]).nonzero()
        ]
        for i, b_idx in enumerate(begin_idxs):
            cur_pred = np.full(cur_pred_tag.shape[0], pred_tag2idx["O"])
            cur_pred[b_idx] = pred_tag2idx["P-B"]
            if i == len(begin_idxs) - 1:
                end_idx = cur_pred_tag.shape[0]
            else:
                end_idx = begin_idxs[i + 1]
            for j, tag in enumerate(cur_pred_tag[b_idx:end_idx]):
                if tag.item() == pred_tag2idx["O"]:
                    break
                elif tag.item() == pred_tag2idx["P-I"]:
                    cur_pred[b_idx + j] = pred_tag2idx["P-I"]
            cur_sent_preds.append(cur_pred)
        if cur_sent_preds:
            total_pred_tags.append(np.vstack(cur_sent_preds))
        else:
            total_pred_tags.append(np.empty(0))
    return [torch.from_numpy(pred_tags) for pred_tags in total_pred_tags]


def get_tuple(sentence, pred_tags, arg_tags, tokenizer):
    """Get string format tuples from given predicate indexes and argument tags.

    :param sentence: string format raw sentence.
    :param pred_tags: tensor of predicate tags with the shape of (# of predicates,
        sequence length).
    :param arg_tags: tensor of argument tags with the same shape.
    :param tokenizer: transformer BertTokenizer (bert-base-cased or
        bert-base-multilingual-cased)

    :return extractions: list of strings each element means predicate, arg0, arg1, ...
    :return extraction_idxs: list of indexes of each argument for calculating
        confidence score.
    """
    word2piece = utils.get_word2piece(sentence, tokenizer)
    words = sentence.split(" ")
    assert pred_tags.shape[0] == arg_tags.shape[0]  # number of predicates

    pred_tags = pred_tags.tolist()
    arg_tags = arg_tags.tolist()
    extractions = list()
    extraction_idxs = list()

    # loop for each predicate
    for cur_pred_tag, cur_arg_tags in zip(pred_tags, arg_tags):
        cur_extraction = list()
        cur_extraction_idxs = list()

        # get predicate
        pred_labels = [pred_tag2idx["P-B"], pred_tag2idx["P-I"]]
        cur_predicate_idxs = [
            idx for idx, tag in enumerate(cur_pred_tag) if tag in pred_labels
        ]
        if len(cur_predicate_idxs) == 0:
            predicates_str = ""
        else:
            cur_pred_words = list()
            for word_idx, piece_idxs in word2piece.items():
                if set(piece_idxs) <= set(cur_predicate_idxs):
                    cur_pred_words.append(word_idx)
            if len(cur_pred_words) == 0:
                predicates_str = ""
                cur_predicate_idxs = list()
            else:
                predicates_str = " ".join([words[idx] for idx in cur_pred_words])
        cur_extraction.append(predicates_str)
        cur_extraction_idxs.append(cur_predicate_idxs)

        # get arguments
        for arg_n in range(4):
            cur_arg_labels = [arg_tag2idx[f"A{arg_n}-B"], arg_tag2idx[f"A{arg_n}-I"]]
            cur_arg_idxs = [
                idx for idx, tag in enumerate(cur_arg_tags) if tag in cur_arg_labels
            ]
            if len(cur_arg_idxs) == 0:
                cur_arg_str = ""
            else:
                cur_arg_words = list()
                for word_idx, piece_idxs in word2piece.items():
                    if set(piece_idxs) <= set(cur_arg_idxs):
                        cur_arg_words.append(word_idx)
                if len(cur_arg_words) == 0:
                    cur_arg_str = ""
                    cur_arg_idxs = list()
                else:
                    cur_arg_str = " ".join([words[idx] for idx in cur_arg_words])
            cur_extraction.append(cur_arg_str)
            cur_extraction_idxs.append(cur_arg_idxs)
        extractions.append(cur_extraction)
        extraction_idxs.append(cur_extraction_idxs)
    return extractions, extraction_idxs


def _find_begins(idxs):
    found_begins = [idxs[0]]
    cur_flag_idx = idxs[0]
    for cur_idx in idxs[1:]:
        if cur_idx - cur_flag_idx != 1:
            found_begins.append(cur_idx)
        cur_flag_idx = cur_idx
    return found_begins


def get_confidence_score(pred_probs, arg_probs, extraction_idxs):
    """Get the confidence score of each extraction for drawing PR-curve.

    :param pred_probs: (sequence length, # of predicate labels)
    :param arg_probs: (# of predicates, sequence length, # of argument labels)
    :param extraction_idxs: [[[2, 3, 4], [0, 1], [9, 10]], [[0, 1, 2], [7, 8],
        [4, 5]], ...]
    """
    confidence_scores = list()
    for cur_arg_prob, cur_ext_idxs in zip(arg_probs, extraction_idxs):
        if len(cur_ext_idxs[0]) == 0:
            confidence_scores.append(0)
            continue
        cur_score = 0

        # predicate score
        pred_score = max(pred_probs[cur_ext_idxs[0][0]]).item()
        cur_score += pred_score

        # argument score
        for arg_idx in cur_ext_idxs[1:]:
            if len(arg_idx) == 0:
                continue
            begin_idxs = _find_begins(arg_idx)
            arg_score = np.mean(
                [max(cur_arg_prob[cur_idx]).item() for cur_idx in begin_idxs],
            )
            cur_score += arg_score
        confidence_scores.append(cur_score)
    return confidence_scores



from .utils import nlp_da  # noqa

from conspiracies.docprocessing.coref import CoreferenceModel


def test_CoreferenceModel(nlp_da):  # noqa
    model = CoreferenceModel()  # check that the model loads as intended

    text = [
        "Hej Kenneth, har du en fed teksts vi kan skrive om dig?",
        "Ja, det kan du tro min fine ven.",
    ]
    docs = nlp_da.pipe(text)

    # test batches forward
    outputs = model.predict_batch_docs(docs)

    # test output format
    for output in outputs:
        assert isinstance(output, dict)



from conspiracies.visualization.graph import (
    get_nodes_edges,
    create_network_graph,
    spring_layout,
)

# Twitter
twitter_week_1_nodes, twitter_week_1_edges = get_nodes_edges(
    "extracted_triplets_tweets/covid_week_1",
    "paraphrase_dim=40_neigh=15_clust=5_samp=3_nodes_edges.json",
    n_most_frequent=3,
)

# GReens
twitter_week_1_graph = create_network_graph(
    twitter_week_1_nodes,
    twitter_week_1_edges,
    title="Twitter (GPT-3): First week of the lockdown",
    k=2.5,
    edge_quantile_value=0.9,
    save="fig/twitter_week_1",
)


# # No få
# twitter_week_1_nodes_rm_få, twitter_week_1_edges_rm_få = get_nodes_edges(
#     "extracted_triplets_tweets/covid_week_1",
#     "paraphrase_dim=40_neigh=15_clust=5_samp=3_nodes_edges.json",
#     remove_custom_nodes=["få"],
#     hard_filter=True,
# )
# twitter_week_1_rm_få = create_network_graph(
#     twitter_week_1_nodes_rm_få,
#     twitter_week_1_edges_rm_få,
#     title="Covid-19 lockdown week 1 - Twitter",
#     k=2.5,
#     node_color="#A82800",
#     fontsize=11,
#     save="fig/twitter_week_1_graph_rm_få",
# )


# With old triplet extraction instead of GPT
twitter_week_1_nodes_multi, twitter_week_1_edges_multi = get_nodes_edges(
    "extracted_triplets_tweets/covid_week_1_multi",
    "paraphrase_dim=40_neigh=15_clust=5_samp=3_nodes_edges.json",
)

twitter_week_1_graph_multi = create_network_graph(
    twitter_week_1_nodes_multi,
    twitter_week_1_edges_multi,
    title="Twitter (Multi2OIE): First week of the lockdown",
    k=3.5,
    edge_quantile_value=0.8,
    save="fig/twitter_week_1_multi_no_labels",
    draw_labels=False,
)

# Mink start
twitter_mink_start_nodes, twitter_mink_start_edges = get_nodes_edges(
    "extracted_triplets_tweets/mink_start",
    "paraphrase_dim=40_neigh=15_clust=5_samp=3_nodes_edges.json",
    hard_filter=True,
)

twitter_mink_start_graph = create_network_graph(
    twitter_mink_start_nodes,
    twitter_mink_start_edges,
    title="Twitter (GPT-3): First week of the mink case",
    layout=spring_layout,
    k=4,
    edge_quantile_value=0.83,
    save="fig/twitter_mink_start",
)

twitter_mink_start_nodes_multi, twitter_mink_start_edges_multi = get_nodes_edges(
    "extracted_triplets_tweets/mink_start_multi",
    "paraphrase_dim=40_neigh=15_clust=5_samp=3_nodes_edges.json",
)

twitter_mink_start_graph_multi = create_network_graph(
    twitter_mink_start_nodes_multi,
    twitter_mink_start_edges_multi,
    title="Twitter (Multi2OIE): First week of the mink case",
    k=3,
    save="fig/twitter_mink_start_multi",
)

### News papers

# Mink start

news_mink_start_nodes, news_mink_start_edges = get_nodes_edges(
    "extracted_triplets_papers/mink_start",
    "paraphrase_dim=40_neigh=15_clust=5_samp=3_nodes_edges.json",
    hard_filter=True,
)

news_mink_start_graph = create_network_graph(
    news_mink_start_nodes,
    news_mink_start_edges,
    title="Newspapers: First week of the mink case",
    layout=spring_layout,
    k=3,
    edge_quantile_value=0.83,
    save="fig/news_mink_start",
)

# mink - Mogens Jensen resigning

news_mink_mj_nodes, news_mink_mj_edges = get_nodes_edges(
    "extracted_triplets_papers/mink_mogens_jensen",
    "paraphrase_dim=40_neigh=15_clust=5_samp=3_nodes_edges.json",
    hard_filter=True,
)

news_mink_mj_graph = create_network_graph(
    news_mink_mj_nodes,
    news_mink_mj_edges,
    title="Newspapers: Mogens Jensen's resignation",
    layout=spring_layout,
    k=2,
    save="fig/news_mink_mj",
)

# Covid week 1

news_week_1_nodes, news_week_1_edges = get_nodes_edges(
    "extracted_triplets_papers/covid_week_1",
    "paraphrase_dim=40_neigh=15_clust=5_samp=3_nodes_edges.json",
)

news_week_2_nodes, news_week_2_edges = get_nodes_edges(
    "extracted_triplets_papers/covid_week_2",
    "paraphrase_dim=40_neigh=15_clust=5_samp=3_nodes_edges.json",
)


news_week_1_graph = create_network_graph(
    news_week_1_nodes,
    news_week_1_edges,
    title="Newspapers: First week of the lockdown",
    save="fig/news_week_1",
    k=2.5,
)

news_week_2_graph = create_network_graph(
    news_week_2_nodes,
    news_week_2_edges,
    title="Newspapers: Second week of the lockdown",
    k=2.5,
    save="fig/news_week_2",
)



import time
import os
import random
from pathlib import Path
from typing import List, Optional, Generator, Union

from spacy.tokens import Doc, Span
import argparse

from data import load_gold_triplets
import spacy
from extract_examples import extract_examples
from conspiracies.docprocessing.relationextraction import (
    MarkdownPromptTemplate2,
    PromptTemplate,
)
import openai

# Conspiracies

from extract_utils import write_txt, ndjson_gen
from src.concat_split_contexts import (
    concat_context,
    tweet_from_context_text,
)


def build_coref_pipeline():
    nlp_coref = spacy.blank("da")
    nlp_coref.add_pipe("sentencizer")
    nlp_coref.add_pipe("allennlp_coref")

    return nlp_coref


def build_headword_extraction_pipeline():
    nlp = spacy.load("da_core_news_sm")
    nlp.add_pipe("sentencizer")
    nlp.add_pipe(
        "heads_extraction",
        config={"normalize_to_entity": True, "normalize_to_noun_chunk": True},
    )
    return nlp


def batch_generator(generator: Generator, n: int):
    """Batches a generator into batches of size n.

    Args:
        generator (Generator): Generator to batch
        n (int): Size of batches

    Yields:
        batch (list): List of elements from generator
    """
    batch = []
    for element in generator:
        batch.append(element)
        if len(batch) == n:
            yield batch
            batch = []
    if batch:
        yield batch


def prepare_template(template: PromptTemplate, n_examples=9, cv=1):
    """Prepares a template by extracting examples from gold docs Uses the
    targets as examples since they are criteria balanced.

    Args:
        template (PromptTemplate): Template to prepare
        n_examples (int, optional): Number of examples to use for prompting.
            Defaults to 10.
        cv (int, optional): Number of times to do the extraction.
            Defaults to 1, since this is only relevant for prompt comparing

    Returns:
        PromptTemplate: the template with examples
    """
    gold_docs = load_gold_triplets(nlp=spacy.load("da_core_news_sm"))
    targets, _ = extract_examples(gold_docs, n_examples, cv)
    return template(examples=targets[0])


def concat_resolve_unconcat_contexts(file_path: str):
    """Concatenates and resolves coreferences in contexts. The resolved
    contexts are then split, and the last (target) tweet is returned.

    Args:
        file_path (str): path to where the contexts are stored

    Returns:
        Generator: a generator with the resolved target tweets
    """
    context_tweets: List[str] = []
    for context in ndjson_gen(file_path):
        concatenated = concat_context(context)
        context_tweets.append(concatenated)

    coref_nlp = build_coref_pipeline()
    coref_docs = coref_nlp.pipe(context_tweets)
    resolved_docs = (d._.resolve_coref for d in coref_docs)

    resolved_tweets = (tweet_from_context_text(tweet) for tweet in resolved_docs)
    return resolved_tweets


def concatenate_tweets(file_path: str, save_path: Optional[str] = None):
    """Concatenates tweets in a file.

    Args:
        file_path (str): path to where the contexts are stored

    Returns:
        Generator: a generator with the concatenated tweets
    """
    context_tweets: List[str] = []
    for context in ndjson_gen(file_path):
        concatenated = concat_context(context)
        context_tweets.append(concatenated)
        if save_path:
            write_txt(save_path, [context[-1]["id"]], "a+")

    return context_tweets


def extract_save_triplets_gpt(
    responses: dict,
    template: PromptTemplate,
    event: str,
    nlp: spacy.Language,
):
    subjects, predicates, objects, triplets = [], [], [], []
    subjects_full, predicates_full, objects_full, triplets_full = [], [], [], []
    for response in responses["choices"]:
        for triplet in template.parse_prompt(response["text"], target_tweet=""):
            if "" in (triplet.subject, triplet.predicate, triplet.object):
                continue
            # Extract elements
            subject = nlp(triplet.subject)
            predicate = nlp(triplet.predicate)
            obj = nlp(triplet.object)

            # Save non-headword reduced
            subjects_full.append(triplet.subject)
            predicates_full.append(triplet.predicate)
            objects_full.append(triplet.object)
            triplets_full.append(
                f"{triplet.subject}, {triplet.predicate}, {triplet.object}",
            )

            # Save headword reduced
            subjects.append(subject._.most_common_ancestor.text)
            predicates.append(predicate._.most_common_ancestor.text)
            objects.append(obj._.most_common_ancestor.text)
            triplets.append(
                f"{subject._.most_common_ancestor.text}, {predicate._.most_common_ancestor.text}, {obj._.most_common_ancestor.text}",
            )

    write_txt(
        os.path.join("extracted_triplets_tweets", event, "subjects.txt"),
        subjects,
        "a+",
    )
    write_txt(
        os.path.join("extracted_triplets_tweets", event, "predicates.txt"),
        predicates,
        "a+",
    )
    write_txt(
        os.path.join("extracted_triplets_tweets", event, "objects.txt"),
        objects,
        "a+",
    )
    write_txt(
        os.path.join("extracted_triplets_tweets", event, "triplets.txt"),
        triplets,
        "a+",
    )
    write_txt(
        os.path.join("extracted_triplets_tweets", event, "subjects_full.txt"),
        subjects_full,
        "a+",
    )
    write_txt(
        os.path.join("extracted_triplets_tweets", event, "predicates_full.txt"),
        predicates_full,
        "a+",
    )
    write_txt(
        os.path.join("extracted_triplets_tweets", event, "objects_full.txt"),
        objects_full,
        "a+",
    )
    write_txt(
        os.path.join("extracted_triplets_tweets", event, "triplets_full.txt"),
        triplets_full,
        "a+",
    )


def prompt_gpt3(
    concatenated_tweets: Generator,
    api_key: Union[str, None],
    event: str,
    batch_size: int = 20,
    org_id: Optional[str] = None,
):
    """Prompts GPT-3 with a template and a generator of concatenated tweets.
    The tweets are resolved with coreference resolution, and then the template
    is used to create a prompt for each tweet. The prompts are then sent to
    GPT-3, and the responses are parsed and saved.

    Args:
        template (PromptTemplate): Template to use for prompting
        concatenated_tweets (Generator): Generator of concatenated tweets
        api_key (str): API key for GPT-3
        event (str): Event name
        batch_size (int, optional): Number of tweets to send in each batch.
            Defaults to 20.
        org_id (Optional[str], optional): Organization ID for GPT-3. Defaults to None.
    """
    assert api_key, "Please provide an API key when using GPT-3"

    # Setup
    print("Loading gold docs and setting up template")
    template = prepare_template(MarkdownPromptTemplate2)
    coref_nlp = build_coref_pipeline()
    head_nlp = build_headword_extraction_pipeline()
    if org_id:
        openai.org_id = org_id
    openai.api_key = api_key
    # Check if the folder for the event exists, if not create it
    Path(os.path.join("extracted_triplets_tweets", event)).mkdir(
        parents=True,
        exist_ok=True,
    )

    print("batching")
    for i, batch in enumerate(batch_generator(concatenated_tweets, batch_size)):
        start = time.time()
        coref_docs = coref_nlp.pipe(batch)
        resolved_docs = (d._.resolve_coref for d in coref_docs)
        resolved_target_tweets = (
            tweet_from_context_text(tweet) for tweet in resolved_docs
        )
        prompts = [
            template.create_prompt(target=tweet) for tweet in resolved_target_tweets
        ]

        print(f"sending request for batch {i}")
        while True:
            try:
                responses = openai.Completion.create(
                    model="text-davinci-002",
                    prompt=prompts,
                    temperature=0.7,
                    max_tokens=500,
                )

                print("parsing response and saving")
                extract_save_triplets_gpt(responses, template, event, head_nlp)
                print(f"batch {i} done in {time.time() - start} seconds\n")
                break

            except openai.error.InvalidRequestError as e:
                print("Invalid request got error: ", e)
                print("Retrying with fewer examples...")
                # Randomly select an example to drop
                current_examples: List[Doc] = list(template.examples)
                current_examples.pop(random.randrange(len(current_examples)))
                template.set_examples(current_examples)  # type: ignore
                prompts = [
                    template.create_prompt(target=tweet)
                    for tweet in resolved_target_tweets
                ]

            except openai.error.APIConnectionError:
                print("Connection reset, waiting 20 sec then retrying...")
                time.sleep(20)
            except openai.error.APIError:
                print("API error, waiting 20 sec then retrying...")
                time.sleep(20)
                continue
            except openai.error.RateLimitError:
                print("RateLimitError, waiting 20 sec then retrying...")
                time.sleep(20)
                continue


def multi2oie_extraction(
    concatenated_tweets: list,
    event: str,
):
    # Check if the folder for the event exists, if not create it
    Path(os.path.join("extracted_triplets_tweets", f"{event}_multi")).mkdir(
        parents=True,
        exist_ok=True,
    )
    print("Building pipeline")
    nlp = spacy.load("da_core_news_sm")
    nlp.add_pipe("sentencizer")
    nlp.add_pipe(
        "heads_extraction",
        config={"normalize_to_entity": True, "normalize_to_noun_chunk": True},
    )
    config = {"confidence_threshold": 2.7, "model_args": {"batch_size": 10}}
    nlp.add_pipe("relation_extractor", config=config)

    docs = nlp.pipe(concatenated_tweets)
    start = time.time()
    i = 0
    run = True
    while run:
        try:
            i += 1
            doc = next(docs)
        except KeyError:
            print("Received a KeyError, skipping tweet")
            docs = nlp.pipe(concatenated_tweets[i:])
            continue
        except StopIteration:
            print("Stopping iteration because of StopIteration exception")
            # TODO: the last iteration happens twice with this logic
            run = False
        subjects, predicates, objects, triplets = [], [], [], []
        for triplet in doc._.relation_triplets:
            if not all(isinstance(element, Span) for element in triplet):
                continue
            if len(triplet) != 3:
                continue
            subject = triplet[0]._.most_common_ancestor.text
            predicate = triplet[1]._.most_common_ancestor.text
            obj = triplet[2]._.most_common_ancestor.text
            subjects.append(subject)
            predicates.append(predicate)
            objects.append(obj)
            triplets.append((subject, predicate, obj))

        if len(triplets) > 0:
            write_txt(
                os.path.join(
                    "extracted_triplets_tweets",
                    f"{event}_multi",
                    "subjects.txt",
                ),
                subjects,
                "a+",
            )
            write_txt(
                os.path.join(
                    "extracted_triplets_tweets",
                    f"{event}_multi",
                    "predicates.txt",
                ),
                predicates,
                "a+",
            )
            write_txt(
                os.path.join(
                    "extracted_triplets_tweets",
                    f"{event}_multi",
                    "objects.txt",
                ),
                objects,
                "a+",
            )
            write_txt(
                os.path.join(
                    "extracted_triplets_tweets",
                    f"{event}_multi",
                    "triplets.txt",
                ),
                triplets,
                "a+",
            )
        if i % 20 == 0 and i != 0:
            print(
                f"Tweet {i} done. Processed 20 tweets in {time.time() - start} seconds\n",
            )
            start = time.time()


def main(
    file_path: str,
    event: str,
    api_key: Optional[str] = None,
    template_batch_size: int = 20,
    sample_size: Union[None, int] = None,
    org_id: Optional[str] = None,
    extraction_method: str = "gpt3",
):
    """Main function for extracting triplets from tweets Uses GPT3 prompting.

    Args:
        file_path (str): path to where the contexts are stored
        event (str): name of the event. Used when saving the extracted triplets
        template_batch_size (int, optional): Number of tweets to prompt GPT3 with at a time.
            Defaults to 20 (max length for list of prompts).
    """
    print("Concatenating tweets")
    concatenated = concatenate_tweets(file_path, f"processed_tweets_{file_path}.txt")

    # Downsampling
    if sample_size:
        if len(concatenated) < sample_size:
            print(
                f"Sample size ({sample_size}) is larger than the number of tweets ({len(concatenated)}), using all tweets",
            )
        else:
            concatenated = random.sample(concatenated, sample_size)
            print(f"Downsampled to {len(concatenated)} tweets")

    print(f"Extracting triplets with {extraction_method}")
    if extraction_method == "gpt3":
        prompt_gpt3(
            concatenated,
            api_key,
            event,
            template_batch_size,
            org_id,
        )
    elif extraction_method == "Multi2OIE":
        multi2oie_extraction(
            concatenated,
            event,
        )
    else:
        raise ValueError(
            f"Extraction method {extraction_method} not supported. Please choose between 'gpt3' and 'Multi2OIE'.",
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-p",
        "--file_path",
        type=str,
        help="Path to the file containing the context tweets",
    )
    parser.add_argument(
        "-e",
        "--event",
        type=str,
        help="Name of the event. Used when saving the extracted triplets",
    )
    parser.add_argument(
        "-t",
        "--template_batch_size",
        type=int,
        default=20,
        required=False,
        help="Number of tweets to prompt GPT3 with at a time. Defaults to 20 (max length for list of prompts).",
    )
    parser.add_argument(
        "-s",
        "--sample_size",
        type=int,
        default=30000,
        required=False,
        help="Number of tweets to downsample dataset to. Defaults to 30000.",
    )
    parser.add_argument(
        "-o",
        "--org_id",
        type=str,
        required=False,
        default=None,
        help="OpenAI organization ID. Defaults to None.",
    )
    parser.add_argument(
        "-api_key",
        "--api_key",
        type=str,
        required=False,
        default=None,
        help="OpenAI API key. Defaults to None.",
    )
    parser.add_argument(
        "-x",
        "--extraction_method",
        type=str,
        required=False,
        default="gpt3",
        help="Extraction method. Defaults to gpt3. Other option is 'Multi2OIE'",
    )
    # file_path = os.path.join("src", "TESTnew_tweet_threads_2019-03-10_2019-03-17.ndjson")
    # event="covid_week_1"
    # template_batch_size=20
    args = parser.parse_args()
    main(
        file_path=args.file_path,
        event=args.event,
        api_key=args.api_key,
        template_batch_size=args.template_batch_size,
        sample_size=args.sample_size,
        org_id=args.org_id,
        extraction_method=args.extraction_method,
    )



from .utils import nlp_en  # noqa

from spacy.tokens import Span

import conspiracies  # noqa


def test_extentions(nlp_en):  # noqa
    nlp_en.add_pipe("heads_extraction")

    doc = nlp_en("Mette Frederiksen is the Danish Politician")
    doc.set_ents([Span(doc, 0, 2, "PERSON")])

    assert isinstance(doc._.most_common_ancestor, Span)  # Doc
    assert isinstance(doc[0:2]._.most_common_ancestor, Span)  # Span



import logging
from typing import Iterator, Iterable

import ndjson

from conspiracies.document import Document


class Preprocessor:
    def __init__(self, metadata_fields: Iterable[str] = ("*",)):
        """
        Args:
            metadata_fields: fields in metadata dict to retain. If '*' is given,
                all fields will be retained.
        """
        self.metadata_fields = set(metadata_fields)

    def _do_preprocess_docs(self, input_path: str) -> Iterator[Document]:
        pass

    def _filter_metadata(
        self,
        preprocessed_docs: Iterator[Document],
    ) -> Iterator[Document]:
        if "*" in self.metadata_fields:
            for doc in preprocessed_docs:
                yield doc

        for doc in preprocessed_docs:
            metadata = doc["metadata"]
            for key in list(metadata.keys()):
                if key not in self.metadata_fields:
                    del metadata[key]
            yield doc

    def _validate_content(
        self,
        preprocessed_docs: Iterator[Document],
    ) -> Iterator[Document]:
        for doc in preprocessed_docs:
            if not doc["text"]:
                logging.warning(
                    "Skipping doc with id '%s' because of empty text field",
                    doc["id"],
                )
                continue
            else:
                yield doc

    def preprocess_docs(self, input_path: str, output_path: str):
        preprocessed_docs = self._do_preprocess_docs(input_path)
        validated = self._validate_content(preprocessed_docs)
        metadata_filtered = self._filter_metadata(validated)
        with open(output_path, "w+") as out_file:
            ndjson.dump(metadata_filtered, out_file)



import pytest
import spacy
from spacy import Vocab

from conspiracies.docprocessing.relationextraction.gptprompting import (
    DocTriplets,
    SpanTriplet,
    StringTriplet,
)
from conspiracies.docprocessing.relationextraction import data_classes
from spacy.tokens import Doc


def test_relationextraction_doc_extension():
    """Verifies the behavior of the relation triplet extension and the lambdas
    that back it."""
    data_classes.install_extensions(force=True)

    words = "this is a test . the test seems cool".split()
    vocab = Vocab(strings=words)
    test_doc = Doc(vocab, words=words)

    this = test_doc[0:1]
    is_ = test_doc[1:2]
    a_test = test_doc[2:4]
    the_test = test_doc[5:7]
    seems = test_doc[7:8]
    cool = test_doc[8:9]

    triplets = DocTriplets(
        span_triplets=[
            SpanTriplet.from_tuple((this, is_, a_test)),
            SpanTriplet.from_tuple((the_test, seems, cool)),
        ],
        doc=test_doc,
    )

    test_doc._.relation_triplets = triplets

    # check setter/getter mirroring
    assert test_doc._.relation_triplets == triplets

    # check index extension
    assert test_doc._.relation_triplet_idxs == [
        ((0, 1), (1, 2), (2, 4)),
        ((5, 7), (7, 8), (8, 9)),
    ]

    # check heads, relations and tails
    assert test_doc._.relation_head == [this, the_test]
    assert test_doc._.relation_relation == [is_, seems]
    assert test_doc._.relation_tail == [a_test, cool]


@pytest.fixture
def nlp():
    nlp = spacy.blank("en")
    nlp.add_pipe("sentencizer")
    return nlp


@pytest.fixture
def doc(nlp):
    return nlp("I am happy today")


@pytest.fixture
def span_triplets(doc):
    triplet = SpanTriplet(
        subject=doc[0:1],
        predicate=doc[1:2],
        object=doc[2:3],
    )

    triplet_w_overlap = SpanTriplet(
        subject=doc[0:2],
        predicate=doc[1:2],
        object=doc[2:3],
    )

    return [triplet, triplet_w_overlap]


@pytest.fixture
def string_triplets(doc):
    text = doc.text

    triplet = StringTriplet(
        subject="I",
        predicate="am",
        object="happy",
        text=text,
    )

    triplet_w_overlap = StringTriplet(
        subject="I am",
        predicate="am",
        object="happy",
        text=text,
    )

    return [triplet, triplet_w_overlap]


class TestSpanTriplet:
    def test_visualize(self, span_triplets):
        for triplet in span_triplets:
            triplet.visualize()

    def test_equal(self, span_triplets):
        triplet1, triplet2 = span_triplets
        assert triplet1 != triplet2
        assert triplet1 == triplet1

    def test_to_from_dict(self, span_triplets, nlp):
        for triplet in span_triplets:
            d = triplet.to_dict()
            _triplet = SpanTriplet.from_dict(d, nlp)
            assert triplet.is_string_match(_triplet)

    @pytest.mark.parametrize(
        "text,str_triplet",
        [
            ("jeg er glad i dag", ["jeg", "er", "glad"]),
            ("jeg er glad i dag", ["jeg", "er", "glad i dag"]),
            ("Kenneth er glad i   dag", ["Kenneth", "er", "glad i   dag"]),
            ("Kenneth er glad i dag", ["Kenneth", "er", "glad i dag"]),
        ],
    )
    def test_from_doc(self, nlp, text, str_triplet):
        doc = nlp(text)
        triplet = SpanTriplet.from_doc(str_triplet, doc=doc, nlp=nlp)

        assert triplet.subject.text == str_triplet[0]
        assert triplet.predicate.text == str_triplet[1]
        assert triplet.object.text == str_triplet[2]


class TestDocTriplet:
    def test_init(self, span_triplets):
        doc = span_triplets[0].doc
        # test empty triplets
        doc_triplets_empty = DocTriplets(span_triplets=[], doc=doc)
        assert len(doc_triplets_empty) == 0
        doc_triplets = DocTriplets(span_triplets=span_triplets, doc=doc)
        assert len(doc_triplets) == 2

        # test add
        new_doc_triplets = doc_triplets_empty + doc_triplets
        assert new_doc_triplets == doc_triplets

        doc_triplets_ = [doc_triplets_empty, doc_triplets, new_doc_triplets]
        for dt in doc_triplets_:
            assert isinstance(dt.doc, Doc)

    def test_doc_triplet_from_str_triplets(self, string_triplets, doc: Doc):
        doc_triplets = DocTriplets.from_str_triplets(triplets=string_triplets, doc=doc)
        assert len(doc_triplets) == 2

    def test_score_relations(self, nlp: spacy.Language):
        doc = nlp("I am happy today I am happy today")

        triplet = SpanTriplet(
            subject=doc[0:1],
            predicate=doc[1:2],
            object=doc[2:3],
        )

        # exact match
        doc_triplets_gold = DocTriplets(span_triplets=[triplet], doc=doc)
        doc_triplets_pred = DocTriplets(span_triplets=[triplet], doc=doc)

        scores = doc_triplets_pred.score_relations(doc_triplets_gold)
        scores["exact_span_match"] = 1
        scores["exact_text_match"] = 1
        scores["normalized_span_overlap"] = 1
        scores["normalized_char_overlap"] = 1

        # exact string match
        triplet_string_match = SpanTriplet(
            subject=doc[4:5],
            predicate=doc[5:6],
            object=doc[6:7],
        )

        doc_triplets_pred = DocTriplets(span_triplets=[triplet_string_match], doc=doc)
        scores = doc_triplets_pred.score_relations(doc_triplets_gold)
        scores["exact_span_match"] = 0
        scores["exact_text_match"] = 1
        scores["normalized_span_overlap"] = 0
        scores["normalized_char_overlap"] = 1

        # partial span match
        triplet_part_span = SpanTriplet(
            subject=doc[0:1],
            predicate=doc[1:2],
            object=doc[2:4],
        )

        doc_triplets_pred = DocTriplets(span_triplets=[triplet_part_span], doc=doc)
        scores = doc_triplets_pred.score_relations(doc_triplets_gold)
        scores["exact_span_match"] = 0
        scores["exact_text_match"] = 0
        scores["normalized_span_overlap"] = 0
        scores["normalized_char_overlap"] = 0.8

        # partial string match
        triplet_part_char = SpanTriplet(
            subject=doc[4:5],
            predicate=doc[5:6],
            object=doc[6:8],
        )

        doc_triplets_pred = DocTriplets(span_triplets=[triplet_part_char], doc=doc)
        scores = doc_triplets_pred.score_relations(doc_triplets_gold)
        scores["exact_span_match"] = 0
        scores["exact_text_match"] = 0
        scores["normalized_span_overlap"] = 0
        scores["normalized_char_overlap"] = 0.8



from spacy.tokens import Span

import conspiracies  # noqa F401

from .utils import nlp_en  # noqa F401


def test_normalize_to_span(nlp_en):  # noqa F811
    nlp_en.add_pipe(
        "heads_extraction",
        config={
            "force": True,
            "normalize_to_entity": False,
            "normalize_to_noun_chunk": False,
        },
    )

    doc = nlp_en("Mette Frederiksen is the Danish politician.")

    normalized_token = doc[0]._.to_span  # Token

    assert isinstance(normalized_token, Span)
    assert normalized_token.text == "Mette"

    normalized_token = doc[0:2]._.to_span  # Span

    assert isinstance(normalized_token, Span)
    assert normalized_token.text == "Mette Frederiksen"

    normalized_token = doc._.to_span  # Doc

    assert isinstance(normalized_token, Span)
    assert normalized_token.text == "Mette Frederiksen is the Danish politician."


def test_normalize_to_entity(nlp_en):  # noqa F811
    nlp_en.add_pipe(
        "heads_extraction",
        config={
            "normalize_to_entity": True,
            "normalize_to_noun_chunk": False,
            "force": True,
        },
    )

    doc = nlp_en("Mette Frederiksen is the Danish politician.")

    normalized_token = doc[0]._.to_span

    assert isinstance(normalized_token, Span)
    assert normalized_token.text == "Mette Frederiksen"


def test_normalize_to_noun_chunk(nlp_en):  # noqa F811
    nlp_en.add_pipe(
        "heads_extraction",
        config={
            "normalize_to_noun_chunk": True,
            "normalize_to_entity": False,
            "force": True,
        },
    )

    doc = nlp_en("Mette Frederiksen is the Danish politician.")

    noun_chunk = doc[1]._.to_span

    assert isinstance(noun_chunk, Span)
    assert noun_chunk.text == "Mette Frederiksen"
    assert noun_chunk.text == "Mette Frederiksen"



from conspiracies.preprocessing.infomedia import InfoMediaPreprocessor


def test_create_text():
    input_text = {
        "Heading": "Nice title",
        "SubHeading": "With a nice subtitle",
        "BodyText": "<p>One paragraph</p>"
        "<p>Another paragraph</p>"
        "<p><br/>A third with a line break and <em>emphasis</em>.</p>",
    }
    text = InfoMediaPreprocessor.create_text(input_text)

    expected = (
        "Nice title\n\n"
        "With a nice subtitle\n\n"
        "One paragraph\n"
        "Another paragraph\n\n"
        "A third with a line break and emphasis."
    )

    assert text == expected
